"""SMPC AST Creation.

__init__ file for SMPC. This defines various modules, classes and methods which we currently support.
We create an AST for all these modules, classes and methods so that they can be called remotely.
"""
# stdlib
from typing import Optional

# relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..node.abstract.node import AbstractNodeClient
from .store import CryptoStore

from .protocol import beaver  # noqa: 401 isort: skip


def create_smpc_ast(client: Optional[AbstractNodeClient] = None) -> Globals:
    """Creates SMPC Abstract Syntax Tree (AST).
        Define a set of modules, classes and methods which are used for the creation of the SMPC AST.

    Args:
        client (Optional[AbstractNodeClient]):  Input client object for AST operations.

    Returns:
        The constructed SMPC AST.
    """
    ast = Globals(client)

    modules = [
        "syft",
        "syft.core",
        "syft.core.smpc",
        "syft.core.smpc.store",
        "syft.core.smpc.protocol",
        "syft.core.smpc.protocol.spdz",
        "syft.core.smpc.protocol.spdz.spdz",
    ]

    classes = [
        (
            "syft.core.smpc.store.CryptoStore",
            "syft.core.smpc.store.CryptoStore",
            CryptoStore,
        ),
    ]

    methods = [
        (
            "syft.core.smpc.store.CryptoStore.get_primitives_from_store",
            "syft.lib.python.List",
        ),
        ("syft.core.smpc.store.CryptoStore.store", "syft.lib.python.Dict"),
        (
            "syft.core.smpc.store.CryptoStore.populate_store",
            "syft.lib.python._SyNone",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pydantic import BaseSettings

# relative
from ....logger import traceback_and_raise
from ...common.message import SignedEventualSyftMessageWithoutReply
from ...common.message import SignedImmediateSyftMessageWithReply
from ...common.message import SignedImmediateSyftMessageWithoutReply
from ...common.uid import UID
from ...io.address import Address
from ...io.location import Location
from ...store import ObjectStore


class AbstractNodeClient(Address):
    lib_ast: Any  # Can't import Globals (circular reference)
    # TODO: remove hacky in_memory_client_registry
    in_memory_client_registry: Dict[Any, Any]
    signing_key: SigningKey
    verify_key: VerifyKey
    """"""

    @property
    def id(self) -> UID:
        """This client points to an node, this returns the id of that node."""
        traceback_and_raise(NotImplementedError)

    def send_immediate_msg_without_reply(self, msg: Any) -> Any:
        raise NotImplementedError

    def send_immediate_msg_with_reply(
        self, msg: Any, timeout: Optional[float] = None, return_signed: bool = False
    ) -> Any:
        raise NotImplementedError


class AbstractNode(Address):

    name: Optional[str]
    signing_key: Optional[SigningKey]
    verify_key: Optional[VerifyKey]
    root_verify_key: VerifyKey
    guest_verify_key_registry: Set[VerifyKey]
    admin_verify_key_registry: Set[VerifyKey]
    cpl_ofcr_verify_key_registry: Set[VerifyKey]
    acc: Optional[Any]
    settings: BaseSettings

    # TODO: remove hacky in_memory_client_registry
    in_memory_client_registry: Dict[Any, Any]
    # TODO: remove hacky signaling_msgs when SyftMessages become Storable.
    signaling_msgs: Dict[Any, Any]

    def __init__(
        self,
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        settings: Optional[BaseSettings] = None,
    ):
        super().__init__(
            name=name, network=network, domain=domain, device=device, vm=vm
        )
        self.settings = settings

    store: ObjectStore
    requests: List
    lib_ast: Any  # Can't import Globals (circular reference)
    """"""

    @property
    def known_child_nodes(self) -> List[Any]:
        traceback_and_raise(NotImplementedError)

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        traceback_and_raise(NotImplementedError)

    @property
    def id(self) -> UID:
        """This client points to an node, this returns the id of that node."""
        traceback_and_raise(NotImplementedError)

    @property
    def keys(self) -> str:
        verify = (
            self.key_emoji(key=self.signing_key.verify_key)
            if self.signing_key is not None
            else "🚫"
        )
        root = (
            self.key_emoji(key=self.root_verify_key)
            if self.root_verify_key is not None
            else "🚫"
        )
        keys = f"🔑 {verify}" + f"🗝 {root}"

        return keys

    def get_peer_client(
        self, node_id: UID, only_vpn: bool = True
    ) -> Optional[AbstractNodeClient]:
        traceback_and_raise(NotImplementedError)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/device.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/torch/device.proto\x12\x0esyft.lib.torch"%\n\x06\x44\x65vice\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x05\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.lib.torch.device_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _DEVICE._serialized_start = 48
    _DEVICE._serialized_end = 85
# @@protoc_insertion_point(module_scope)



"""CryptoStore manages the needed crypto primitives."""

# stdlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List

# relative
from ...common.serde.serializable import serializable


@serializable(recursive_serde=True)
class CryptoStore:
    """Manages items needed for MPC Computation.

    Attributes:
        store (Dict[Any, Any]): Object needed for MPC.

    """

    __slots__ = ("store",)
    __attr_allowlist__ = ("store",)

    _func_add_store: Dict[Any, Callable] = {}
    _func_get_store: Dict[Any, Callable] = {}

    def __init__(self) -> None:
        """Initializer."""
        self.store: Dict[Any, Any] = {}

    def populate_store(
        self,
        op_str: str,
        primitives: Iterable[Any],
        *args: List[Any],
        **kwargs: Dict[Any, Any]
    ) -> None:
        """Populate items.

        Args:
            op_str (str): Operator to store.
            primitives (Iterables[Any]): Primitives to store.
            *args: Variable length arguments.
            **kwargs: Keywords arguments.
        """
        populate_func = CryptoStore._func_add_store[op_str]
        populate_func(self.store, primitives, *args, **kwargs)

    def get_primitives_from_store(
        self, op_str: str, ring_size: int, *args: List[Any], **kwargs: Dict[Any, Any]
    ) -> List[Any]:
        """Get primitives from store.

        Args:
            op_str (str): Operator to get.
            nr_instances (int): Number of instances.
            *args: Variable length arguments.
            **kwargs: Keywords arguments.

        Returns:
            Dict[Any, Any]: Primitives.
        """
        retrieve_func = CryptoStore._func_get_store[op_str]
        primitives = retrieve_func(
            store=self.store, ring_size=ring_size, *args, **kwargs
        )

        return primitives



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......util import traceback_and_raise
from ....abstract.node_service_interface import NodeServiceInterface
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .vpn_messages import VPNConnectMessage
from .vpn_messages import VPNConnectMessageWithReply
from .vpn_messages import VPNConnectReplyMessage
from .vpn_messages import VPNJoinMessage
from .vpn_messages import VPNJoinMessageWithReply
from .vpn_messages import VPNJoinReplyMessage
from .vpn_messages import VPNJoinSelfMessage
from .vpn_messages import VPNJoinSelfMessageWithReply
from .vpn_messages import VPNJoinSelfReplyMessage
from .vpn_messages import VPNRegisterMessage
from .vpn_messages import VPNRegisterMessageWithReply
from .vpn_messages import VPNRegisterReplyMessage
from .vpn_messages import VPNStatusMessage
from .vpn_messages import VPNStatusMessageWithReply
from .vpn_messages import VPNStatusReplyMessage


class VPNConnectService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: VPNConnectMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> VPNConnectReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process VPNConnectService with no verification key."
            )

        result = msg.payload.run(node=node, verify_key=verify_key)
        return VPNConnectMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[VPNConnectMessage]]:
        return [VPNConnectMessage]


class VPNJoinService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: VPNJoinMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> VPNJoinReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process VPNJoinService with no verification key."
            )

        result = msg.payload.run(node=node, verify_key=verify_key)
        return VPNJoinMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[VPNJoinMessage]]:
        return [VPNJoinMessage]


class VPNJoinSelfService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: VPNJoinSelfMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> VPNJoinSelfReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process VPNJoinService with no verification key."
            )

        result = msg.payload.run(node=node, verify_key=verify_key)
        return VPNJoinSelfMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[VPNJoinSelfMessage]]:
        return [VPNJoinSelfMessage]


class VPNRegisterService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: VPNRegisterMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> VPNRegisterReplyMessage:
        # this service requires no verify_key because its currently public
        result = msg.payload.run(node=node)
        return VPNRegisterMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[VPNRegisterMessage]]:
        return [VPNRegisterMessage]


class VPNStatusService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: VPNStatusMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> VPNStatusReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process VPNJoinService with no verification key."
            )

        result = msg.payload.run(node=node)
        return VPNStatusMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[VPNStatusMessage]]:
        return [VPNStatusMessage]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/store/store_object.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# third party
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2

# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n#proto/core/store/store_object.proto\x12\x0fsyft.core.store\x1a%proto/core/common/common_object.proto\x1a\x19google/protobuf/any.proto"\x82\x01\n\x12StoreGenericObject\x12\x38\n\x0fstorable_object\x18\x01 \x01(\x0b\x32\x1f.syft.core.store.StorableObject\x12\x32\n\x0cstore_object\x18\x02 \x01(\x0b\x32\x1c.syft.core.store.StoreObject"\xfa\x01\n\x0eStorableObject\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x11\n\tdata_type\x18\x02 \x01(\t\x12\x1a\n\x12schematic_qualname\x18\x03 \x01(\t\x12"\n\x04\x64\x61ta\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x18\n\x10read_permissions\x18\x07 \x01(\x0c\x12\x1a\n\x12search_permissions\x18\x08 \x01(\x0c\x12\x19\n\x11write_permissions\x18\t \x01(\x0c"r\n\x0bStoreObject\x12-\n\tcontainer\x18\x01 \x01(\x0b\x32\x1a.syft.core.store.Container\x12\x34\n\rsearch_engine\x18\x02 \x01(\x0b\x32\x1d.syft.core.store.SearchEngine"\xb1\x01\n\tContainer\x12\x36\n\x0e\x64ict_container\x18\x01 \x01(\x0b\x32\x1e.syft.core.store.DictContainer\x12\x34\n\rsql_container\x18\x02 \x01(\x0b\x32\x1d.syft.core.store.SqlContainer\x12\x36\n\x0egrid_container\x18\x03 \x01(\x0b\x32\x1e.syft.core.store.GridContainer"\x0e\n\x0cSearchEngine"\xa0\x01\n\rDictContainer\x12\x45\n\x0c\x64ict_mapping\x18\x01 \x03(\x0b\x32/.syft.core.store.DictContainer.DictMappingEntry\x1aH\n\x10\x44ictMappingEntry\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x38\x01"\x0e\n\x0cSqlContainer"\x0f\n\rGridContainerb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.store.store_object_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _DICTCONTAINER_DICTMAPPINGENTRY._options = None
    _DICTCONTAINER_DICTMAPPINGENTRY._serialized_options = b"8\001"
    _STOREGENERICOBJECT._serialized_start = 123
    _STOREGENERICOBJECT._serialized_end = 253
    _STORABLEOBJECT._serialized_start = 256
    _STORABLEOBJECT._serialized_end = 506
    _STOREOBJECT._serialized_start = 508
    _STOREOBJECT._serialized_end = 622
    _CONTAINER._serialized_start = 625
    _CONTAINER._serialized_end = 802
    _SEARCHENGINE._serialized_start = 804
    _SEARCHENGINE._serialized_end = 818
    _DICTCONTAINER._serialized_start = 821
    _DICTCONTAINER._serialized_end = 981
    _DICTCONTAINER_DICTMAPPINGENTRY._serialized_start = 909
    _DICTCONTAINER_DICTMAPPINGENTRY._serialized_end = 981
    _SQLCONTAINER._serialized_start = 983
    _SQLCONTAINER._serialized_end = 997
    _GRIDCONTAINER._serialized_start = 999
    _GRIDCONTAINER._serialized_end = 1014
# @@protoc_insertion_point(module_scope)



"""This module contains `Global`, an AST node representing the collection
of frameworks held in the global namespace."""

# stdlib
from types import ModuleType
from typing import Any
from typing import Callable as CallableT
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# relative
from ..core.common.uid import UID
from ..logger import traceback_and_raise
from .attribute import Attribute
from .callable import Callable
from .module import Module


class Globals(Module):
    """The collection of frameworks held in the global namespace."""

    registered_clients: Dict[UID, Any] = {}
    loaded_lib_constructors: Dict[str, CallableT] = {}

    def __init__(
        self,
        client: Optional[Any],
        object_ref: Optional[Union[CallableT, ModuleType]] = None,
        parent: Optional[Attribute] = None,
        path_and_name: Optional[str] = None,
        return_type_name: Optional[str] = None,
    ):
        """Base constructor for all frameworks in global namespace.

        Args:
            client: The client who will use all frameworks for executing all the computations.
            object_ref: The python object for which computation is being performed.
            parent: The parent node is optional as importing frameworks don't need any parent node.
            path_and_name: The path for current frameworks in global namespace, e.g. `syft.lib.python.List`.
            return_type_name: The return type name of given action as a string with it's full path.
        """
        super().__init__(
            client=client,
            object_ref=object_ref,
            parent=parent,
            path_and_name=path_and_name,
            return_type_name=return_type_name,
        )

    def __call__(
        self,
        path: Union[List[str], str],
        index: int = 0,
        obj_type: Optional[type] = None,
    ) -> Optional[Union[Callable, CallableT]]:
        """Execute the given node object reference with the framework paths in global space.

        Args:
            path: The node path in module to execute, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List"].
            index : The associated position in the path for the current node.
            obj_type: The type of the object to be called, whose path is resolved from `lookup_cache`.

        Returns:
            The results of running the computation on the object reference.
        """
        self.apply_node_changes()

        _path: List[str] = (
            path.split(".") if isinstance(path, str) else path if path else []
        )

        if not _path:
            traceback_and_raise(
                ValueError("Can't execute remote call if path is not specified.")
            )

        return self.attrs[_path[index]](path=_path, index=index + 1, obj_type=obj_type)

    def add_path(
        self,
        path: Union[str, List[str]] = None,  # type:  ignore
        index: int = 0,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        """Adds new nodes in the AST based on the type of the current node and the type of the object to be added.

        Args:
            path: The node path added in the AST, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List"].
            index: The associated position in the path for the current node.
            framework_reference: The python framework in which we can resolve the same path to obtain the Python object.
            return_type_name: The return type name of given action as a string with it's full path.
            is_static: If the queried object is static, it has to found on the ast itself, not on an existing pointer.
        """
        if isinstance(path, str):
            path = path.split(".")

        framework_name = path[index]

        if framework_name not in self.attrs:
            if framework_reference is None:
                traceback_and_raise(
                    Exception(
                        "You must pass in a framework object, the first time you add method \
                        within the framework."
                    )
                )

            self.attrs[framework_name] = Module(
                path_and_name=".".join(path[: index + 1]),
                object_ref=framework_reference,
                return_type_name=return_type_name,
                client=self.client,
                parent=self,
            )

        attr = self.attrs[framework_name]
        attr.add_path(path=path, index=1, return_type_name=return_type_name)

    def register_updates(self, client: Any) -> None:
        """Any previously loaded libs need to be applied.

        Args:
            client: The client who will be executing all the computations.
        """
        for _, update_ast in self.loaded_lib_constructors.items():
            update_ast(ast_or_client=client)

        """Make sure to get any future updates."""
        self.registered_clients[client.id] = client

    def apply_node_changes(self) -> None:
        """Global is not expected to change."""
        pass



# stdlib
from typing import Tuple
from typing import Union

# third party
from capnp.lib.capnp import _DynamicStructBuilder
import numpy as np
import pyarrow as pa
import torch

# relative
from ...core.common.serde.capnp import CapnpModule
from ...core.common.serde.capnp import chunk_bytes
from ...core.common.serde.capnp import combine_bytes
from ...core.common.serde.capnp import get_capnp_schema
from ...core.common.serde.serializable import serializable
from ...experimental_flags import ApacheArrowCompression
from ...experimental_flags import flags
from ...proto.lib.numpy.array_pb2 import NumpyProto
from ..torch.tensor_util import tensor_deserializer
from ..torch.tensor_util import tensor_serializer

SUPPORTED_BOOL_TYPES = [np.bool_]
SUPPORTED_INT_TYPES = [
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]

SUPPORTED_FLOAT_TYPES = [
    np.float16,
    np.float32,
    np.float64,
]

SUPPORTED_DTYPES = SUPPORTED_BOOL_TYPES + SUPPORTED_INT_TYPES + SUPPORTED_FLOAT_TYPES

DTYPE_REFACTOR = {
    np.dtype("uint16"): np.int16,
    np.dtype("uint32"): np.int32,
    np.dtype("uint64"): np.int64,
}


# TODO: move to sy.serialize interface, when protobuf for numpy is removed.
def capnp_serialize(obj: np.ndarray, to_bytes: bool = False) -> _DynamicStructBuilder:
    schema = get_capnp_schema(schema_file="array.capnp")
    array_struct: CapnpModule = schema.Array  # type: ignore
    array_msg = array_struct.new_message()
    metadata_schema = array_struct.TensorMetadata
    array_metadata = metadata_schema.new_message()

    obj_bytes, obj_decompressed_size = arrow_serialize(obj, get_bytes=True)
    chunk_bytes(obj_bytes, "array", array_msg)
    array_metadata.dtype = str(obj.dtype)
    array_metadata.decompressedSize = obj_decompressed_size

    array_msg.arrayMetadata = array_metadata

    if not to_bytes:
        return array_msg
    else:
        return array_msg.to_bytes_packed()


def capnp_deserialize(
    msg: Union[_DynamicStructBuilder, bytes], from_bytes: bool = False
) -> np.ndarray:
    array_msg: _DynamicStructBuilder
    if from_bytes:
        schema = get_capnp_schema(schema_file="array.capnp")
        array_struct: CapnpModule = schema.Array  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        # to pack or not to pack?
        # array_msg = array_struct.from_bytes(buf, traversal_limit_in_words=2 ** 64 - 1)
        array_msg = array_struct.from_bytes_packed(
            msg, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )
    else:
        array_msg = msg

    array_metadata = array_msg.arrayMetadata
    obj = arrow_deserialize(
        combine_bytes(array_msg.array),
        decompressed_size=array_metadata.decompressedSize,
        dtype=array_metadata.dtype,
    )

    return obj


def arrow_serialize(
    obj: np.ndarray, get_bytes: bool = False
) -> Union[Tuple[bytes, int], NumpyProto]:
    original_dtype = obj.dtype
    apache_arrow = pa.Tensor.from_numpy(obj=obj)
    sink = pa.BufferOutputStream()
    pa.ipc.write_tensor(apache_arrow, sink)
    buffer = sink.getvalue()
    if flags.APACHE_ARROW_COMPRESSION is ApacheArrowCompression.NONE:
        numpy_bytes = buffer.to_pybytes()
    else:
        numpy_bytes = pa.compress(
            buffer, asbytes=True, codec=flags.APACHE_ARROW_COMPRESSION.value
        )
    dtype = original_dtype.name

    if get_bytes:
        return numpy_bytes, buffer.size
    else:
        return NumpyProto(
            arrow_data=numpy_bytes, dtype=dtype, decompressed_size=buffer.size
        )


def arrow_deserialize(buf: bytes, decompressed_size: int, dtype: str) -> np.ndarray:
    original_dtype = np.dtype(dtype)
    if flags.APACHE_ARROW_COMPRESSION is ApacheArrowCompression.NONE:
        reader = pa.BufferReader(buf)
        buf = reader.read_buffer()
    else:
        buf = pa.decompress(
            buf,
            decompressed_size=decompressed_size,
            codec=flags.APACHE_ARROW_COMPRESSION.value,
        )

    result = pa.ipc.read_tensor(buf)
    np_array = result.to_numpy()
    np_array.setflags(write=True)
    return np_array.astype(original_dtype)


def protobuf_serialize(obj: np.ndarray) -> NumpyProto:
    original_dtype = obj.dtype
    if original_dtype not in SUPPORTED_DTYPES:
        raise NotImplementedError(f"{original_dtype} is not supported")

    if original_dtype in DTYPE_REFACTOR:
        # store as a signed int, the negative wrap around values convert back to the
        # same original unsigned values on the other side
        obj = obj.astype(DTYPE_REFACTOR[original_dtype])

    # Cloning seems to cause the worker to freeze if the array is larger than around
    # 800k in data and since we are serializing it immediately afterwards I don't
    # think its needed anyway
    # tensor = torch.from_numpy(obj).clone()
    tensor = torch.from_numpy(obj)
    tensor_bytes = tensor_serializer(tensor)
    dtype = original_dtype.name
    return NumpyProto(proto_data=tensor_bytes, dtype=dtype)


def protobuf_deserialize(proto: NumpyProto) -> np.ndarray:
    tensor = tensor_deserializer(proto.proto_data)
    array = tensor.to("cpu").detach().numpy().copy()
    str_dtype = proto.dtype
    original_dtype = np.dtype(str_dtype)
    obj = array.astype(original_dtype)
    return obj


def serialize_numpy_array(obj: np.ndarray) -> NumpyProto:
    if flags.APACHE_ARROW_TENSOR_SERDE:
        return arrow_serialize(obj)
    else:
        return protobuf_serialize(obj)


def deserialize_numpy_array(proto: NumpyProto) -> np.ndarray:
    if proto.HasField("arrow_data"):
        return arrow_deserialize(
            buf=proto.arrow_data,
            decompressed_size=proto.decompressed_size,
            dtype=proto.dtype,
        )
    else:
        return protobuf_deserialize(proto)


serializable(generate_wrapper=True)(
    wrapped_type=np.ndarray,
    import_path="numpy.ndarray",
    protobuf_scheme=NumpyProto,
    type_object2proto=serialize_numpy_array,
    type_proto2object=deserialize_numpy_array,
)



# stdlib
from typing import Any

# relative
from ...core.common import UID


class PyPrimitive:
    def __init__(self, temporary_box: bool = False) -> None:
        self._id: UID

        # sometimes we need to serialize a python primitive in such a way that we can
        # deserialize it back as that primitive later. This flag allows us to do that.
        self.temporary_box = temporary_box

    def upcast(self) -> Any:
        pass



# stdlib
from typing import Tuple

# third party
import gevent

# relative
from .....logger import critical
from ....common.uid import UID
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode


def retrieve_object(
    node: AbstractNode, id_at_location: UID, path: str, proxy_only: bool = False
) -> StorableObject:
    # A hard time limit is set on celery worker which prevents infinite execution.
    ctr = 0
    while True:
        store_obj = node.store.get_or_none(key=id_at_location, proxy_only=proxy_only)
        if store_obj is None:
            if ctr % 1500 == 0:
                critical(
                    f"execute_action on {path} failed due to missing object"
                    + f" at: {id_at_location}"
                )
            # Implicit context switch between greenlets.
            gevent.sleep(0)
            ctr += 1
        else:
            return store_obj


def beaver_retrieve_object(
    node: AbstractNode, id_at_location: UID, nr_parties: int
) -> StorableObject:
    # relative
    from .beaver_action import BEAVER_CACHE

    # A hard time limit is set on celery worker which prevents infinite execution.
    ctr = 0
    while True:
        store_obj = BEAVER_CACHE.get(id_at_location, None)  # type: ignore
        if store_obj is None or len(store_obj.data) != nr_parties:
            if ctr % 1500 == 0:
                critical(
                    f"Beaver Retrieval failed for {nr_parties} parties due to missing object"
                    + f" at: {id_at_location} values: {store_obj}"
                )
            # Implicit context switch between greenlets.
            gevent.sleep(0)
            ctr += 1
        else:
            return store_obj


def crypto_store_retrieve_object(  # type: ignore
    op_str: str,
    **kwargs,
) -> Tuple:
    # relative
    from ....smpc.store.exceptions import EmptyPrimitiveStore
    from ....tensor.smpc.share_tensor import ShareTensor

    crypto_store = ShareTensor.crypto_store

    # A hard time limit is set on celery worker which prevents infinite execution.
    ctr = 0
    while True:
        try:
            store_values = crypto_store.get_primitives_from_store(
                op_str, **kwargs  # type: ignore
            )
            return tuple(store_values)
        except EmptyPrimitiveStore:

            if ctr % 1500 == 0:
                critical(
                    f"Crypto Store Retrieval failed for parties due to missing object: {EmptyPrimitiveStore}"
                )
            # Implicit context switch between greenlets.
            gevent.sleep(0)
            ctr += 1



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/action.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.node.common.action import (
    action_sequence_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_action__sequence__pb2,
)
from syft.proto.core.node.common.action import (
    garbage_collect_object_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_garbage__collect__object__pb2,
)
from syft.proto.core.node.common.action import (
    get_enum_attribute_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__enum__attribute__pb2,
)
from syft.proto.core.node.common.action import (
    get_object_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__object__pb2,
)
from syft.proto.core.node.common.action import (
    get_set_property_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__property__pb2,
)
from syft.proto.core.node.common.action import (
    get_set_static_attribute_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_get__set__static__attribute__pb2,
)
from syft.proto.core.node.common.action import (
    run_class_method_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_run__class__method__pb2,
)
from syft.proto.core.node.common.action import (
    run_function_or_constructor_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_run__function__or__constructor__pb2,
)
from syft.proto.core.node.common.action import (
    save_object_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_save__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/core/node/common/action/action.proto\x12\x1csyft.core.node.common.action\x1a.proto/core/node/common/action/get_object.proto\x1a?proto/core/node/common/action/run_function_or_constructor.proto\x1a\x34proto/core/node/common/action/run_class_method.proto\x1a:proto/core/node/common/action/garbage_collect_object.proto\x1a\x36proto/core/node/common/action/get_enum_attribute.proto\x1a\x34proto/core/node/common/action/get_set_property.proto\x1a<proto/core/node/common/action/get_set_static_attribute.proto\x1a/proto/core/node/common/action/save_object.proto\x1a\x33proto/core/node/common/action/action_sequence.proto"\xc6\x06\n\x06\x41\x63tion\x12\x10\n\x08obj_type\x18\x01 \x01(\t\x12J\n\x11get_object_action\x18\x02 \x01(\x0b\x32-.syft.core.node.common.action.GetObjectActionH\x00\x12j\n"run_function_or_constructor_action\x18\x03 \x01(\x0b\x32<.syft.core.node.common.action.RunFunctionOrConstructorActionH\x00\x12U\n\x17run_class_method_action\x18\x04 \x01(\x0b\x32\x32.syft.core.node.common.action.RunClassMethodActionH\x00\x12\x61\n\x1dgarbage_collect_object_action\x18\x06 \x01(\x0b\x32\x38.syft.core.node.common.action.GarbageCollectObjectActionH\x00\x12U\n\x15\x65num_attribute_action\x18\x07 \x01(\x0b\x32\x34.syft.core.node.common.action.GetEnumAttributeActionH\x00\x12Z\n\x1aget_or_set_property_action\x18\x08 \x01(\x0b\x32\x34.syft.core.node.common.action.GetOrSetPropertyActionH\x00\x12\x64\n\x1fget_set_static_attribute_action\x18\t \x01(\x0b\x32\x39.syft.core.node.common.action.GetSetStaticAttributeActionH\x00\x12L\n\x12save_object_action\x18\n \x01(\x0b\x32..syft.core.node.common.action.SaveObjectActionH\x00\x12G\n\x0f\x61\x63tion_sequence\x18\x0b \x01(\x0b\x32,.syft.core.node.common.action.ActionSequenceH\x00\x42\x08\n\x06\x61\x63tionb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.node.common.action.action_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ACTION._serialized_start = 578
    _ACTION._serialized_end = 1416
# @@protoc_insertion_point(module_scope)



"""This file defines the class to process the request to regenerate
the type pointer corresponding to the current pointer."""


# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ....abstract.node import AbstractNode
from ..node_service import ImmediateNodeServiceWithReply
from .resolve_pointer_type_messages import ResolvePointerTypeAnswerMessage
from .resolve_pointer_type_messages import ResolvePointerTypeMessage


class ResolvePointerTypeService(ImmediateNodeServiceWithReply):
    """Validate the current pointer and return the corresponding type resolved pointer."""

    @staticmethod
    def process(
        node: AbstractNode,
        msg: ResolvePointerTypeMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> ResolvePointerTypeAnswerMessage:
        """Return the type pointer corresponding to the current pointer.

        Args:
            node (AbstractNode): node client.
            msg (ResolvePointerTypeMessage): address of the pointer whose type needs to be resolved.
            verify_key (Optional[VerifyKey], optional): digital signature of the user. Defaults to None.

        Returns:
            ResolvePointerTypeAnswerMessage: details of the generated pointer type.
        """
        # TODO: refactor so we can get the pointer without deserializing the whole
        # object which could be over in the blob store
        object = node.store.get(msg.id_at_location, proxy_only=False)
        type_qualname = object.object_qualname
        return ResolvePointerTypeAnswerMessage(
            address=msg.reply_to, type_path=type_qualname
        )

    @staticmethod
    def message_handler_types() -> List[Type[ResolvePointerTypeMessage]]:
        return [ResolvePointerTypeMessage]



# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ...core.common import UID
from ...core.common.serde.serializable import serializable
from ...proto.lib.python.bool_pb2 import Bool as Bool_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


def dispatch_other(obj: Any) -> bool:
    if isinstance(obj, Bool):
        return obj.value
    return obj


@serializable()
class Bool(int, PyPrimitive):
    def __new__(cls, value: Any = None, id: Optional[UID] = None) -> "Bool":
        value = bool(value)
        obj = int.__new__(cls, value)
        return obj

    def __init__(self, value: Any = None, id: Optional[UID] = None):
        self.value: bool = bool(value)
        self._id: UID = id if id else UID()
        self.my_field: int = 0

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.
        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def upcast(self) -> bool:
        return bool(self)

    def __abs__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__abs__())

    def __add__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__add__(other))

    def __and__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__and__(other))

    def __bool__(self) -> bool:
        return bool(self.value)

    def __ceil__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__ceil__())

    def __divmod__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        tpl = self.value.__divmod__(other)
        return PrimitiveFactory.generate_primitive(value=tpl)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__eq__(other))

    def __float__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__float__())

    def __floor__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__floor__())

    def __floordiv__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__floordiv__(other))

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__ge__(other))

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__gt__(other))

    def __hash__(self) -> int:
        return PrimitiveFactory.generate_primitive(value=self.value.__hash__())

    def __invert__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__invert__())

    def __int__(self) -> int:
        return int(self.value)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__le__(other))

    def __lshift__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__lshift__(other))

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__lt__(other))

    def __mod__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__mod__(other))

    def __mul__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__mul__(other))

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__ne__(other))

    def __neg__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__neg__())

    def __or__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__or__(other))

    def __pos__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__pos__())

    def __pow__(self, other: Any, modulo: Optional[Any] = None) -> SyPrimitiveRet:
        if modulo:
            PrimitiveFactory.generate_primitive(value=super().__pow__(other, modulo))
        return PrimitiveFactory.generate_primitive(value=super().__pow__(other))

    def __radd__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__radd__(other))

    def __rand__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rand__(other))

    def __repr__(self) -> str:
        return bool(self.value).__repr__()

    def __rdivmod__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        tpl = self.value.__rdivmod__(other)

        return PrimitiveFactory.generate_primitive(value=tpl)

    def __rfloordiv__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(
            value=self.value.__rfloordiv__(other)
        )

    def __rlshift__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rlshift__(other))

    def __rmod__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rmod__(other))

    def __rmul__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rmul__(other))

    def __ror__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__ror__(other))

    def __round__(self, ndigits: Any = 0) -> SyPrimitiveRet:
        dispatch_ndigits = dispatch_other(ndigits)
        return PrimitiveFactory.generate_primitive(
            value=self.value.__round__(dispatch_ndigits)
        )

    def __rpow__(self, other: Any, modulo: Optional[Any] = None) -> SyPrimitiveRet:
        other = dispatch_other(other)

        if modulo:
            return PrimitiveFactory.generate_primitive(
                value=self.value.__rpow__(other, modulo)
            )
        return PrimitiveFactory.generate_primitive(value=self.value.__rpow__(other))

    def __rrshift__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rrshift__(other))

    def __rshift__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rshift__(other))

    def __rsub__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rsub__(other))

    def __rtruediv__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rtruediv__(other))

    def __rxor__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__rxor__(other))

    def __str__(self) -> str:
        return bool(self.value).__str__()

    def __sub__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__sub__(other))

    def __truediv__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__truediv__(other))

    def __trunc__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.__trunc__())

    def __xor__(self, other: Any) -> SyPrimitiveRet:
        other = dispatch_other(other)
        return PrimitiveFactory.generate_primitive(value=self.value.__xor__(other))

    def as_integer_ratio(self) -> SyPrimitiveRet:
        res = self.value.as_integer_ratio()
        return PrimitiveFactory.generate_primitive(value=res)

    def bit_length(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.bit_length())

    def conjugate(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.conjugate())

    def denominator(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.denominator)

    # TODO: add support for properties on these 4 functions

    def imag(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.imag)

    def numerator(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.numerator)

    def real(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=self.value.real)

    def _object2proto(self) -> Bool_PB:
        return Bool_PB(id=sy.serialize(obj=self.id), data=self)

    @staticmethod
    def _proto2object(proto: Bool_PB) -> "Bool":
        return Bool(id=sy.deserialize(blob=proto.id), value=proto.data)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Bool_PB



# stdlib
from typing import Optional

# third party
from google.protobuf.message import Message
from google.protobuf.reflection import GeneratedProtocolMessageType

# relative
from ....logger import traceback_and_raise
from ....util import random_name
from ...common.uid import UID


class Location:
    """This represents the location of a node, including
    location-relevant metadata (such as how long it takes
    for us to communicate with this location, etc.)"""

    def __init__(self, name: Optional[str] = None) -> None:
        if name is None:
            name = random_name()
        self.name = name
        super().__init__()

    @property
    def id(self) -> UID:
        traceback_and_raise(NotImplementedError)

    def repr_short(self) -> str:
        """Returns a SHORT human-readable version of the ID

        Return a SHORT human-readable version of the ID which
        makes it print nicer when embedded (often alongside other
        UID objects) within other object __repr__ methods."""

        return self.__repr__()

    @staticmethod
    def _proto2object(proto: Message) -> "Location":
        """This method converts a protobuf object into a subclass of Serializable

        This method must be implemented for all classes which subclassSerializable - namely
        all classes which can be serialized within the Syft ecosystem. It should convert the
        corresponding protobuf message for the subclass into an instance of the class. This
        allows all the logic which goes from protobuf message to other formats (JSON, binary, etc.)
        to be generic and simply inherited from the rest of this class.

        :param proto: the protobuf object to be converted into an instance of type(self)
        :param type: Message
        :return: an instance of type(self)
        :rtype: Serializable

        """
        traceback_and_raise(NotImplementedError)

    def _object2proto(self) -> Message:
        """This methods converts self into a protobuf object

        This method must be implemented by all subclasses so that generic high-level functions
        implemented here (such as serialize(, to_bytes=True), etc) know how to convert the object into
        a protobuf object before further converting it into the requested format.

        :return: a protobuf message
        :rtype: Message
        """

        traceback_and_raise(NotImplementedError)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        traceback_and_raise(NotImplementedError)



# DOs and Don's of this class:
# - Do NOT use absolute syft imports (i.e. import syft.core...) Use relative ones.
# - Do NOT put multiple imports on the same line (i.e. from <x> import a, b, c). Use separate lines
# - Do sort imports by length
# - Do group imports by where they come from

# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey
from typing_extensions import final

# syft absolute
import syft as sy

# relative
from ......proto.core.node.common.service.object_search_permission_update_message_pb2 import (
    ObjectSearchPermissionUpdateMessage as ObjectSearchPermissionUpdateMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
@final
class ObjectSearchPermissionUpdateMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        add_instead_of_remove: bool,
        target_verify_key: Optional[VerifyKey],
        target_object_id: UID,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)

        self.add_instead_of_remove = add_instead_of_remove
        self.target_verify_key = target_verify_key
        self.target_object_id = target_object_id

    def _object2proto(self) -> ObjectSearchPermissionUpdateMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectSearchPermissionUpdateMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return ObjectSearchPermissionUpdateMessage_PB(
            msg_id=sy.serialize(self.id),
            address=sy.serialize(self.address),
            target_verify_key=bytes(self.target_verify_key)
            if self.target_verify_key
            else None,
            target_object_id=sy.serialize(self.target_object_id),
            add_instead_of_remove=self.add_instead_of_remove,
        )

    @staticmethod
    def _proto2object(
        proto: ObjectSearchPermissionUpdateMessage_PB,
    ) -> "ObjectSearchPermissionUpdateMessage":
        """Creates a ObjectSearchPermissionUpdateMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of ObjectSearchPermissionUpdateMessage
        :rtype: ObjectSearchPermissionUpdateMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return ObjectSearchPermissionUpdateMessage(
            msg_id=sy.deserialize(blob=proto.msg_id),
            address=sy.deserialize(blob=proto.address),
            target_verify_key=VerifyKey(proto.target_verify_key)
            if proto.target_verify_key
            else None,
            target_object_id=sy.deserialize(blob=proto.target_object_id),
            add_instead_of_remove=proto.add_instead_of_remove,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return ObjectSearchPermissionUpdateMessage_PB



# future
from __future__ import annotations

# stdlib
import functools
import itertools
import secrets
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union

# third party
import numpy as np
import numpy.typing as npt
import torch

# relative
from . import utils
from .... import logger
from ....grid import GridURL
from ...node.common.action.przs_action import PRZSAction
from ...smpc.approximations import APPROXIMATIONS
from ...smpc.protocol.spdz import spdz
from ...smpc.store import CryptoPrimitiveProvider
from ..config import DEFAULT_RING_SIZE
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import SupportedChainType  # type: ignore
from ..util import implements  # type: ignore
from .share_tensor import ShareTensor

if TYPE_CHECKING:
    # relative
    from ..tensor import Tensor
    from ..tensor import TensorPointer

METHODS_FORWARD_ALL_SHARES = {
    "repeat",
    "copy",
    "diagonal",
    "flatten",
    "transpose",
    "partition",
    "resize",
    "ravel",
    "compress",
    "reshape",
    "squeeze",
    "swapaxes",
    "sum",
    "__pos__",
    "__neg__",
    "take",
    "choose",
    "cumsum",
    "trace",
}
INPLACE_OPS = {
    "resize",
}

PARTIES_REGISTER_CACHE: Dict[Any, GridURL] = {}


class MPCTensor(PassthroughTensor):
    __slots__ = (
        "seed_przs",
        "mpc_shape",
        "parties",
        "parties_info",
        "ring_size",
    )

    def __init__(
        self,
        parties: List[Any],
        secret: Optional[Any] = None,
        shares: Optional[Union[List[Tensor], List[TensorPointer]]] = None,
        shape: Optional[Tuple[int, ...]] = None,
        seed_przs: Optional[int] = None,
        ring_size: Optional[int] = None,
    ) -> None:

        if secret is None and shares is None:
            raise ValueError("Secret or shares should be populated!")

        if (shares is not None) and (not isinstance(shares, (tuple, list))):
            raise ValueError("Shares should be a list or tuple")

        if seed_przs is None:
            # Allow the user to specify if they want to use a specific seed when generating the shares
            # ^This is unsecure and should be used with cautioness
            seed_przs = secrets.randbits(32)

        self.seed_przs = seed_przs
        self.parties = parties
        self.parties_info = MPCTensor.get_parties_info(parties)

        if ring_size is not None:
            self.ring_size = ring_size
        else:
            self.ring_size = MPCTensor.get_ring_size_from_secret(secret, shares)
        self.mpc_shape = shape

        # TODO: We can get this from the the secret if the secret is local
        # TODO: https://app.clubhouse.io/openmined/story/1128/tech-debt-for-adp-smpc-demo?stories_sort_by\
        #  =priority&stories_group_by=WORKFLOW_STATE
        if shape is None:
            raise ValueError("Shape of the secret should be known")

        if secret is not None:
            shares = MPCTensor._get_shares_from_secret(
                secret=secret,
                parties=parties,
                parties_info=self.parties_info,
                shape=shape,
                seed_przs=seed_przs,
                ring_size=self.ring_size,
            )

        if shares is None:
            raise ValueError("Shares should not be None at this step")

        res = list(MPCTensor._mpc_from_shares(shares, parties=parties))

        # we need to make sure that when we zip up clients from
        # multiple MPC tensors that they are in corresponding order
        # so we always sort all of them by the id of the domain
        # TODO: store children as lists of dictionaries because eventually
        # it's likely that we have multiple shares from the same client
        # (For example, if you wanted a domain to have 90% share ownership
        # you'd need to produce 10 shares and give 9 of them to the same domain)
        # TODO captured: https://app.clubhouse.io/openmined/story/1128/tech-debt-for-adp-smpc-\
        #  demo?stories_sort_by=priority&stories_group_by=WORKFLOW_STATE
        res.sort(key=lambda share: share.client.name + share.client.id.no_dash)

        super().__init__(res)

    @staticmethod
    def get_ring_size_from_secret(
        secret: Optional[Any] = None, shares: Optional[List[Any]] = None
    ) -> int:
        value = shares[0] if secret is None else secret  # type: ignore

        if utils.ispointer(value):
            dtype = getattr(value, "public_dtype", None)
        else:
            dtype = getattr(value, "dtype", None)

        ring_size: Optional[int] = (
            value.ring_size
            if isinstance(value, MPCTensor)
            else utils.TYPE_TO_RING_SIZE.get(dtype, None)
        )

        if ring_size is not None:
            return ring_size

        logger.warning(f"Ring size was not found! Defaulting to {DEFAULT_RING_SIZE}.")
        return DEFAULT_RING_SIZE

    @staticmethod
    def get_parties_info(parties: Iterable[Any]) -> List[GridURL]:
        # relative
        from ....grid.client import GridHTTPConnection

        parties_info: List[GridURL] = []
        for party in parties:
            connection = party.routes[0].connection
            if not isinstance(connection, GridHTTPConnection):
                raise TypeError(
                    f"You tried to pass {type(connection)} for multiplication dependent operation."
                    + "Currently Syft works only with hagrid"
                    + "We apologize for the inconvenience"
                    + "We will add support for local python objects very soon."
                )
            base_url = PARTIES_REGISTER_CACHE.get(party, None)

            if base_url is None:
                base_url = connection.base_url
                PARTIES_REGISTER_CACHE[party] = base_url
                try:
                    pass
                    # We do not use sy.register, should reenable after fixing.
                    # sy.register(  # nosec
                    #     name="Howard Wolowtiz",
                    #     email="howard@mit.edu",
                    #     password="astronaut",
                    #     url=url,
                    #     port=port,
                    #     verbose=False,
                    # )
                except Exception:
                    """ """
                    # TODO : should modify to return same client if registered.
                    # print("Proxy Client already User Register", e)
            if base_url is not None:
                parties_info.append(base_url)
            else:
                raise Exception(
                    f"Failed to get GridURL from {base_url} for party {party}."
                )

        return parties_info

    def publish(self, sigma: float) -> MPCTensor:
        new_shares = []

        for share in self.child:
            share.block

        for share in self.child:
            new_share = share.publish(sigma=sigma)
            new_shares.append(new_share)

        return MPCTensor(
            parties=self.parties,
            shares=new_shares,
            shape=self.mpc_shape,
            seed_przs=self.seed_przs,
        )

    @property
    def shape(self) -> Optional[Tuple[int, ...]]:
        return self.mpc_shape

    @staticmethod
    def _mpc_from_shares(
        shares: Union[List[Tensor], List[TensorPointer]],
        parties: Optional[List[Any]] = None,
    ) -> List[TensorPointer]:
        if not isinstance(shares, (list, tuple)):
            raise ValueError("_mpc_from_shares expected a list or tuple of shares")

        if utils.ispointer(shares[0]):
            # Remote shares
            return shares  # type: ignore
        elif parties is None:
            raise ValueError(
                "Parties should not be None if shares are not already sent to parties"
            )
        else:
            return MPCTensor._mpc_from_local_shares(shares, parties)  # type: ignore

    @staticmethod
    def _mpc_from_local_shares(
        shares: List[Tensor], parties: List[Any]
    ) -> List[TensorPointer]:
        # TODO: ShareTensor needs to have serde serializer/deserializer
        shares_ptr = [share.send(party) for share, party in zip(shares, parties)]
        return shares_ptr

    @staticmethod
    def _get_shares_from_secret(
        secret: Any,
        parties: List[Any],
        shape: Tuple[int, ...],
        seed_przs: int,
        parties_info: List[GridURL],
        ring_size: int,
    ) -> Union[List[Tensor], List[TensorPointer]]:
        if utils.ispointer(secret):
            if shape is None:
                raise ValueError("Shape must be specified when the secret is remote")
            return MPCTensor._get_shares_from_remote_secret(
                secret=secret,
                shape=shape,
                parties=parties,
                seed_przs=seed_przs,
                parties_info=parties_info,
                ring_size=ring_size,
            )

        return MPCTensor._get_shares_from_local_secret(
            secret=secret,
            seed_przs=seed_przs,
            ring_size=ring_size,
            shape=shape,
            parties_info=parties_info,
        )

    @staticmethod
    def _get_shares_from_remote_secret(
        secret: Any,
        shape: Tuple[int, ...],
        parties: List[Any],
        seed_przs: int,
        parties_info: List[GridURL],
        ring_size: int,
    ) -> List[TensorPointer]:
        shares = []
        for i, party in enumerate(parties):
            if secret is not None and party == secret.client:
                value = secret
            else:
                value = None

            # relative
            from ..autodp.gamma_tensor import TensorWrappedGammaTensorPointer
            from ..autodp.phi_tensor import TensorWrappedPhiTensorPointer

            is_dp_tensor = False
            if isinstance(
                secret, (TensorWrappedPhiTensorPointer, TensorWrappedGammaTensorPointer)
            ):

                share_wrapper = secret.to_local_object_without_private_data_child()

                kwargs = {
                    "rank": i,
                    "parties_info": parties_info,
                    "value": value,
                    "shape": shape,
                    "seed_przs": seed_przs,
                    "share_wrapper": share_wrapper,
                    "ring_size": str(ring_size),
                }
                is_dp_tensor = True
                attr_path_and_name = "syft.core.tensor.smpc.share_tensor.ShareTensor.generate_przs_on_dp_tensor"

            else:
                kwargs = {
                    "rank": i,
                    "parties_info": parties_info,
                    "value": value,
                    "shape": shape,
                    "seed_przs": seed_przs,
                    "ring_size": str(ring_size),
                }
                attr_path_and_name = (
                    "syft.core.tensor.smpc.share_tensor.ShareTensor.generate_przs"
                )

            args: List[Any] = []  # Currently we do not use any args in PRZS Action

            return_type_name = party.lib_ast.query(attr_path_and_name).return_type_name
            resolved_pointer_type = party.lib_ast.query(return_type_name)
            result = resolved_pointer_type.pointer_type(client=party)

            # TODO: Path in PRZS is not used due to attribute error syft ast for przs of ShareTensor,
            # should be modified to use path.
            # QUESTION can the id_at_location be None?
            result_id_at_location = getattr(result, "id_at_location", None)
            if result_id_at_location is not None:
                cmd = PRZSAction(
                    path=attr_path_and_name,
                    args=args,
                    kwargs=kwargs,
                    id_at_location=result_id_at_location,
                    is_dp_tensor=is_dp_tensor,
                    address=party.address,
                )
                party.send_immediate_msg_without_reply(msg=cmd)

            shares.append(result)

        return shares

    @staticmethod
    def _get_shares_from_local_secret(
        secret: Any,
        shape: Tuple[int, ...],
        seed_przs: int,
        parties_info: List[GridURL],
        ring_size: int = DEFAULT_RING_SIZE,
    ) -> List[Tensor]:
        shares = []
        nr_parties = len(parties_info)
        for i in range(nr_parties):
            if i == nr_parties - 1:
                value = secret
            else:
                value = None

            local_share = ShareTensor.generate_przs(
                rank=i,
                parties_info=parties_info,
                value=value,
                shape=shape,
                seed_przs=seed_przs,
                init_clients=False,
                ring_size=ring_size,
            )

            shares.append(local_share)

        return shares

    def request(
        self,
        reason: str = "",
        block: bool = False,
        timeout_secs: Optional[int] = None,
        verbose: bool = False,
    ) -> None:
        for child in self.child:
            child.request(
                reason=reason, block=block, timeout_secs=timeout_secs, verbose=verbose
            )

    @property
    def block(self) -> "MPCTensor":
        """Block until all shares have been created."""
        for share in self.child:
            share.block

        return self

    def block_with_timeout(self, secs: int, secs_per_poll: int = 1) -> "MPCTensor":
        """Block until all shares have been created or until timeout expires."""

        for share in self.child:
            share.block_with_timeout(secs=secs, secs_per_poll=secs_per_poll)

        return self

    def reconstruct(self, delete_obj: bool = True) -> np.ndarray:
        # relative
        from ..fixed_precision_tensor import FixedPrecisionTensor

        dtype = utils.RING_SIZE_TO_TYPE.get(self.ring_size, None)

        if dtype is None:
            raise ValueError(f"Type for ring size {self.ring_size} was not found!")

        for share in self.child:
            if not share.exists:
                raise Exception(
                    "One of the shares doesn't exist. This probably means the SMPC "
                    "computation isn't yet complete. Try again in a moment or call .block.reconstruct()"
                    "instead to block until the SMPC operation is complete which creates this variable."
                )

        local_shares = []
        for share in self.child:
            res = share.get() if delete_obj else share.get_copy()
            local_shares.append(res)

        result = local_shares[0]
        op = ShareTensor.get_op(self.ring_size, "add")
        for share in local_shares[1:]:
            result = op(result, share)

        def check_fpt(value: Any) -> bool:
            if isinstance(value, FixedPrecisionTensor):
                return True
            if hasattr(value, "child"):
                return check_fpt(value.child)
            else:
                return False

        if check_fpt(result):
            return result.decode()

        def get_lowest_child(value: Any) -> AcceptableSimpleType:
            if isinstance(value, PassthroughTensor):
                return get_lowest_child(value.child)
            else:
                return value

        return get_lowest_child(result)

    get = reconstruct

    def get_copy(self) -> np.ndarray:
        return self.reconstruct(delete_obj=False)

    @staticmethod
    def hook_method(__self: MPCTensor, method_name: str) -> Callable[..., Any]:
        """Hook a framework method.

        Args:
            method_name (str): method to hook

        Returns:
            A hooked method
        """

        def method_all_shares(
            _self: MPCTensor, *args: List[Any], **kwargs: Dict[Any, Any]
        ) -> Any:

            shares = []

            for share in _self.child:
                method = getattr(share, method_name)
                new_share = method(*args, **kwargs)
                shares.append(new_share)

                # TODO: generalize type after fixed precision
                dummy_res = np.random.randint(
                    _self.mpc_shape[0], size=_self.mpc_shape, dtype=np.int32  # type: ignore
                )
                if method_name not in INPLACE_OPS:
                    dummy_res = getattr(dummy_res, method_name)(*args, **kwargs)
                else:
                    getattr(dummy_res, method_name)(*args, **kwargs)

                new_shape = dummy_res.shape
            res = MPCTensor(parties=_self.parties, shares=shares, shape=new_shape)
            return res

        return functools.partial(method_all_shares, __self)

    def __getattribute__(self, attr_name: str) -> Any:

        if attr_name in METHODS_FORWARD_ALL_SHARES:
            return MPCTensor.hook_method(self, attr_name)

        return object.__getattribute__(self, attr_name)

    @staticmethod
    def reshare(mpc_tensor: MPCTensor, parties: Iterable[Any]) -> MPCTensor:
        """Reshare a given secret to a superset of parties.
        Args:
            mpc_tensor(MPCTensor): input MPCTensor to reshare.
            parties(List[Any]): Input parties List.
        Returns:
            res_mpc(MPCTensor): Reshared MPCTensor.
        Raises:
            ValueError: If the input MPCTensor and input parties are same.

        Note:
        We provide an additional layer of abstraction such that,
        when computation is performed on data belonging to different parties
        The underlying secret are automatically converted into secret shares of their input.

        Assume there are two parties Parties P1,P2

        tensor_1 = data_pointer_1 (party 1 data)
        tensor_2 = data_pointer_2 (party 2 data)

        result -------> tensor_1+ tensor_1 (local computation as the data
        belongs to the same party)

        Interesting scenario is when

        result --------> tensor_1+tensor_2

        Each tensor belongs to two different parties.
        There are automatically secret shared without the user
        knowing that a MPCTensor is being created underneath.
        """
        mpc_parties = set(mpc_tensor.parties)
        parties = set(parties)
        shape = mpc_tensor.shape
        seed_przs = mpc_tensor.seed_przs
        client_map = {share.client: share for share in mpc_tensor.child}

        if mpc_parties == parties:
            raise ValueError(
                "Input parties for resharing are same as the input parties."
            )
        parties_info = MPCTensor.get_parties_info(parties)
        shares = [client_map.get(party) for party in parties]
        for i, party in enumerate(parties):
            shares[
                i
            ] = party.syft.core.tensor.smpc.share_tensor.ShareTensor.generate_przs(
                rank=i,
                parties_info=parties_info,
                value=shares[i],
                shape=shape,
                seed_przs=seed_przs,
                ring_size=str(mpc_tensor.ring_size),
            )

        res_mpc = MPCTensor(shares=shares, ring_size=mpc_tensor.ring_size, shape=shape, parties=parties)  # type: ignore

        return res_mpc

    @staticmethod
    def sanity_checks(mpc_tensor: MPCTensor, other: Any) -> Tuple[MPCTensor, Any]:
        """Performs sanity checks to share data to whole superset of parites involved.
        Args:
            mpc_tensor(MPCTensor): input MPCTensor to perform sanity check on.
            other (Any): input operand.
        Returns:
            Tuple[MPCTensor,Any]: Rehared Tensor values.
        """
        if utils.ispointer(other):
            parties = mpc_tensor.parties
            client = other.client
            public_shape = other.public_shape
            if public_shape is None:
                # TODO: Should be modified after Trask's Synthetic data PR.
                raise ValueError("The input tensor pointer should have public shape.")
            if client not in parties:
                new_parties = [client]
                new_parties += parties
                mpc_tensor = MPCTensor.reshare(mpc_tensor, new_parties)
            else:
                new_parties = parties

            other = MPCTensor(secret=other, parties=new_parties, shape=public_shape)

        elif isinstance(other, MPCTensor):
            p1 = set(mpc_tensor.parties)  # parties in first MPCTensor
            p2 = set(other.parties)  # parties in second MPCTensor.
            if p1 != p2:
                parties_union = p1.union(p2)
                mpc_tensor = (
                    MPCTensor.reshare(mpc_tensor, parties_union)
                    if p1 != parties_union
                    else mpc_tensor
                )
                other = (
                    MPCTensor.reshare(other, parties_union)
                    if p2 != parties_union
                    else other
                )

        return mpc_tensor, other

    def __apply_private_op(
        self, other: MPCTensor, op_str: str, **kwargs: Dict[Any, Any]
    ) -> List[ShareTensor]:

        op_method = f"__{op_str}__"
        if op_str in {"add", "sub"}:
            if len(self.child) != len(other.child):
                raise ValueError(
                    "Zipping two different lengths will drop data. "
                    + f"{len(self.child)} vs {len(other.child)}"
                )
            res_shares = [
                getattr(a, op_method)(b, **kwargs)
                for a, b in zip(self.child, other.child)
            ]

        else:
            raise ValueError(f"MPCTensor Private {op_str} not supported")
        return res_shares

    def __apply_public_op(
        self, y: Any, op_str: str, **kwargs: Dict[Any, Any]
    ) -> List[ShareTensor]:
        op_method = f"__{op_str}__"
        if op_str in {"mul", "matmul", "add", "sub"}:
            res_shares = [
                getattr(share, op_method)(y, **kwargs) for share in self.child
            ]

        else:
            raise ValueError(f"MPCTensor Public {op_str} not supported")

        return res_shares

    def __apply_op(
        self,
        y: Union[int, float, torch.Tensor, np.ndarray, MPCTensor],
        op_str: str,
    ) -> MPCTensor:
        """Apply an operation on "self" which is a MPCTensor "y".

         This function checks if "y" is private or public value.

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, MPCTensor]: tensor to apply the operation.
            op_str (str): the operation.

        Returns:
            MPCTensor. the operation "op_str" applied on "self" and "y"
        """

        x, y = MPCTensor.sanity_checks(self, y)
        kwargs: Dict[Any, Any] = {"seed_id_locations": secrets.randbits(64)}

        y_ring_size = MPCTensor.get_ring_size_from_secret(y)
        if self.ring_size != y_ring_size:
            raise ValueError(
                f"Ring size mismatch between self {self.ring_size} and other {y_ring_size}"
            )

        if isinstance(y, MPCTensor):
            result = x.__apply_private_op(y, op_str, **kwargs)
        else:
            result = x.__apply_public_op(y, op_str, **kwargs)

        y_shape = getattr(y, "shape", (1,))
        shape = utils.get_shape(op_str, self.shape, y_shape)
        ring_size = utils.get_ring_size(self.ring_size, y_ring_size)

        result = MPCTensor(
            shares=result, shape=shape, ring_size=ring_size, parties=x.parties
        )

        return result

    def add(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        """Apply the "add" operation between "self" and "y".

        Args:
            y (Union[MPCTensor, torch.Tensor, float, int]): self + y

        Returns:
            MPCTensor. Result of the operation.
        """
        res = self.__apply_op(y, "add")
        return res

    def sub(self, y: MPCTensor) -> MPCTensor:
        res = self.__apply_op(y, "sub")
        return res

    def rsub(self, y: MPCTensor) -> MPCTensor:
        new_self = self * (-1)
        res = new_self.__apply_op(y, "add")
        return res

    def mul(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:

        self, y = MPCTensor.sanity_checks(self, y)
        kwargs: Dict[Any, Any] = {"seed_id_locations": secrets.randbits(64)}
        op = "__mul__"
        res_shares: List[Any] = []
        y_shape = getattr(y, "shape", (1,))
        new_shape = utils.get_shape("mul", self.mpc_shape, y_shape)
        if isinstance(y, MPCTensor):
            res_shares = spdz.mul_master(self, y, "mul", **kwargs)
        else:
            CryptoPrimitiveProvider.generate_primitives(
                "beaver_wraps",
                parties=self.parties,
                g_kwargs={
                    "shape": new_shape,
                    "parties_info": self.parties_info,
                },
                p_kwargs={"shape": new_shape},
                ring_size=self.ring_size,
            )
            res_shares = [
                getattr(a, op)(b, **kwargs) for a, b in zip(self.child, itertools.repeat(y))  # type: ignore
            ]

        res = MPCTensor(
            parties=self.parties,
            shares=res_shares,
            shape=new_shape,
            ring_size=self.ring_size,
        )

        return res

    def matmul(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        self, y = MPCTensor.sanity_checks(self, y)
        kwargs: Dict[Any, Any] = {"seed_id_locations": secrets.randbits(64)}
        op = "__matmul__"
        res_shares: List[Any] = []
        y_shape = getattr(y, "shape", (1,))
        new_shape = utils.get_shape("matmul", self.mpc_shape, y_shape)
        if isinstance(y, MPCTensor):
            res_shares = spdz.mul_master(self, y, "matmul", **kwargs)
        else:
            CryptoPrimitiveProvider.generate_primitives(
                "beaver_wraps",
                parties=self.parties,
                g_kwargs={
                    "shape": new_shape,
                    "parties_info": self.parties_info,
                },
                p_kwargs={"shape": new_shape},
                ring_size=self.ring_size,
            )

            res_shares = [
                getattr(a, op)(b, **kwargs) for a, b in zip(self.child, itertools.repeat(y))  # type: ignore
            ]

        res = MPCTensor(
            parties=self.parties,
            shares=res_shares,
            shape=new_shape,
            ring_size=self.ring_size,
        )

        return res

    def lt(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        mpc_res = spdz.lt_master(self, y, "mul")  # type: ignore

        return mpc_res

    def gt(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        mpc_res = MPCTensor.lt(y, self)  # type: ignore

        return mpc_res

    def ge(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:

        mpc_res = 1 - MPCTensor.lt(self, y)

        return mpc_res  # type: ignore

    def le(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        mpc_res = 1 - MPCTensor.lt(y, self)  # type: ignore

        return mpc_res  # type: ignore

    def eq(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        # TODO: Should make two comparisons parallel
        mpc_res = MPCTensor.le(self, y) - MPCTensor.lt(self, y)

        return mpc_res

    def ne(
        self, y: Union[int, float, np.ndarray, torch.tensor, MPCTensor]
    ) -> MPCTensor:
        mpc_res = 1 - MPCTensor.eq(self, y)

        return mpc_res  # type: ignore

    def truediv(self, y: Union["MPCTensor", np.ndarray, float, int]) -> MPCTensor:
        """Apply the "div" operation between "self" and "y".

        Args:
            y (Union["MPCTensor", torch.Tensor, float, int]): Denominator.

        Returns:
            MPCTensor: Result of the operation.

        Raises:
            ValueError: If input denominator is float.
        """
        is_private = isinstance(y, MPCTensor)

        result: MPCTensor
        if is_private:
            reciprocal = APPROXIMATIONS["reciprocal"]
            result = self.mul(reciprocal(y))  # type: ignore
        else:
            result = self * (1 / y)

        return result

    def rtruediv(self, y: Union[np.ndarray, float, int]) -> MPCTensor:
        """Apply recriprocal of MPCTensor.

        Args:
            y (Union[torch.Tensor, float, int]): Numerator.

        Returns:
            MPCTensor: Result of the operation.
        """
        reciprocal = APPROXIMATIONS["reciprocal"]
        result: MPCTensor = reciprocal(self) * y  # type: ignore
        return result

    def put(
        self,
        indices: npt.ArrayLike,
        values: npt.ArrayLike,
        mode: Optional[str] = "raise",
    ) -> MPCTensor:
        """Performs Numpy put operation on the underlying ShareTensors.
        Args:
            indices (npt.ArrayLike): Target indices, interpreted as integers.
            values (npt.ArrayLike): Values to place at target indices.
            mode (Optional[str]): Specifies how out-of-bounds indices will behave.
        Returns:
            res (MPCTensor): Result of the operation.
        """
        shares = []
        shares.append(self.child[0].put(indices, values, mode))
        # since the value is public we assign directly to prevent overhead of random share creation.
        zero = np.zeros_like(values)
        for share in self.child[1::]:
            shares.append(share.put(indices, zero.copy(), mode))

        res = MPCTensor(shares=shares, parties=self.parties, shape=self.shape)
        return res

    def concatenate(
        self,
        other: Union[MPCTensor, TensorPointer],
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ) -> MPCTensor:
        self, other = MPCTensor.sanity_checks(self, other)

        shares = []
        for x, y in zip(self.child, other.child):  # type: ignore
            shares.append(x.concatenate(y, *args, **kwargs))

        dummy_res = np.concatenate(
            (np.empty(self.shape), np.empty(other.shape)), *args, **kwargs  # type: ignore
        )
        res = MPCTensor(shares=shares, parties=self.parties, shape=dummy_res.shape)

        return res

    def sign(self) -> MPCTensor:
        """Calculate sign of given tensor.

        Returns:
            MPCTensor: with computed sign.
        """
        res: MPCTensor = 2 * (self > 0) - 1  # type: ignore
        return res

    def __abs__(self) -> MPCTensor:
        """Calculates absolute value of MPCTensor.

        Returns:
            MPCTensor: computed absolute values of underlying tensor.
        """
        return self * self.sign()

    def __pow__(self, power: int) -> MPCTensor:
        """Compute integer power of a number iteratively using mul.

        - Divide power by 2 and multiply base to itself (if the power is even)
        - Decrement power by 1 to make it even and then follow the first step

        Args:
            power (int): integer value to apply the

        Returns:
             MPCTensor: Result of the pow operation

        Raises:
            RuntimeError: if negative power is given
        """
        # TODO: Implement after we have reciprocal function.
        if power < 0:
            raise RuntimeError("Negative integer powers not supported yet.")

        base = self

        # TODO: should modify for general ring sizes.
        result = np.ones(shape=self.shape, dtype=np.int32)

        while power > 0:
            # If power is odd
            if power % 2 == 1:
                result = base * result
                result.block

            # Divide the power by 2
            power = power // 2
            # Multiply base to itself
            base = base * base
            base.block

        return result

    def __str__(self) -> str:
        res = "MPCTensor"
        for share in self.child:
            res = f"{res}\n\t{share}"

        return res

    @property
    def synthetic(self) -> np.ndarray:
        # TODO finish. max_vals and min_vals not available at present.
        public_dtype_func = getattr(
            self.public_dtype, "upcast", lambda: self.public_dtype
        )
        return (
            np.random.rand(*list(self.shape)) * (self.max_vals - self.min_vals)  # type: ignore
            + self.min_vals
        ).astype(public_dtype_func())

    def __repr__(self) -> str:

        # out = self.synthetic.__repr__()
        # out += "\n\n (The data printed above is synthetic - it's an imitation of the real data.)"
        out = ""
        out += "\n\nMPCTensor"
        out += ".shape=" + str(self.shape) + "\n"
        for i, child in enumerate(self.child):
            out += f"\t .child[{i}] = " + child.__repr__() + "\n"
        out = out[:-1] + ""

        return out

    __add__ = add
    __radd__ = add
    __sub__ = sub
    __rsub__ = rsub
    __mul__ = mul
    __rmul__ = mul
    __matmul__ = matmul
    __lt__ = lt
    __gt__ = gt
    __ge__ = ge
    __le__ = le
    __eq__ = eq
    __ne__ = ne
    __truediv__ = truediv
    __rtruediv__ = rtruediv


@implements(MPCTensor, np.add)
def add(x: np.ndarray, y: MPCTensor) -> SupportedChainType:
    return y.add(x)


@implements(MPCTensor, np.subtract)
def sub(x: np.ndarray, y: MPCTensor) -> SupportedChainType:
    return y.rsub(x)


@implements(MPCTensor, np.multiply)
def mul(x: np.ndarray, y: MPCTensor) -> SupportedChainType:
    return y.mul(x)


# @implements(MPCTensor, np.greater)
# def mul(x: np.ndarray, y: MPCTensor) -> SupportedChainType:
#     return y.gt(x)



# third party
from sqlalchemy import Column
from sqlalchemy import JSON
from sqlalchemy import String

# relative
from . import Base


class Dataset(Base):
    __tablename__ = "dataset"

    id = Column(String(256), primary_key=True)
    name = Column(String(256))
    manifest = Column(String(2048))
    description = Column(String(2048))
    tags = Column(JSON())
    str_metadata = Column(JSON())
    blob_metadata = Column(JSON())



"""
Welcome to the syft package! This package is the primary package for PySyft.
This package has two kinds of attributes: submodules and convenience functions.
Submodules are configured in the standard way, but the convenience
functions exist to allow for a convenient `import syft as sy` to then expose
the most-used functionalities directly on syft. Note that this way of importing
PySyft is the strict convention in this codebase. (Do no simply call
`import syft` and then directly use `syft.<method>`.)
The syft module is split into two distinct groups of functionality which we casually refer to
as syft "core" and syft "python". "core" functionality is functionality which is designed
to be universal across all Syft languages (javascript, kotlin, swift, etc.).
Syft "python" includes all functionality which by its very nature cannot be
truly polyglot. Syft "core" functionality includes the following modules:
* :py:mod:`syft.core.node` - APIs for interacting with remote machines you do not directly control.
* :py:mod:`syft.core.message` - APIs for serializing messages sent between Client and Node classes.
* :py:mod:`syft.core.pointer` - Client side API for referring to objects on a Node
* :py:mod:`syft.core.store` - Server side API for referring to object storage on a node (things pointers point to)
Syft "python" functionality includes the following modules:
* :py:mod:`syft.ast` - code generates external library common syntax tree using an allowlist list of methods
* :py:mod:`syft.typecheck` - automatically checks and enforces Python type hints and the exclusive use of kwargs.
* :py:mod:`syft.lib` - uses the ast library to dynamically create remote execution APIs for supported Python libs.
    IMPORTANT: syft.core should be very careful when importing functionality from outside of syft
    core!!! Since we plan to drop syft core down to a language (such as C++ or Rust)
    this can create future complications with lower level languages calling
    higher level ones.
To begin your education in Syft, continue to the :py:mod:`syft.core.node.vm.vm` module...
"""

__version__ = "0.7.0-beta.41"

# stdlib
from pathlib import Path
import sys
from typing import Any

# third party
from pkg_resources import DistributionNotFound  # noqa: F401
from pkg_resources import get_distribution  # noqa: F401

# relative
# Package Imports
from . import filterwarnings  # noqa: F401
from . import jax_settings  # noqa: F401
from . import lib  # noqa: F401
from . import logger  # noqa: F401

# ASTRACT OBJECT IMPORTS
from .core import common  # noqa: F401
from .core.adp.data_subject_list import DataSubjectArray  # noqa: F401
from .core.adp.data_subject_list import DataSubjectList  # noqa: F401

# Convenience Methods
from .core.common.serde.deserialize import _deserialize as deserialize  # noqa: F401
from .core.common.serde.serialize import _serialize as serialize  # noqa: F401
from .core.node.common.node_service.testing_services.repr_service import (  # noqa: F401
    ReprMessage,
)
from .core.node.device import Device  # noqa: F401
from .core.node.device_client import DeviceClient  # noqa: F401
from .core.node.domain import Domain  # noqa: F401
from .core.node.domain import DomainClient  # noqa: F401
from .core.node.network import Network  # noqa: F401
from .core.node.network_client import NetworkClient  # noqa: F401

# Convenience Constructors
from .core.node.vm import VirtualMachine  # noqa: F401
from .core.node.vm_client import VirtualMachineClient  # noqa: F401
from .core.tensor import autodp  # noqa: F401
from .core.tensor import nn  # noqa: F401
from .core.tensor.autodp.gamma_tensor import GammaTensor  # noqa: F401
from .core.tensor.autodp.phi_tensor import PhiTensor  # noqa: F401
from .core.tensor.lazy_repeat_array import lazyrepeatarray  # noqa: F401
from .core.tensor.tensor import Tensor  # noqa: F401
from .experimental_flags import flags  # noqa: F401
from .grid.client.client import connect  # noqa: F401
from .grid.client.client import login  # noqa: F401
from .grid.client.client import register  # noqa: F401

# Convenience Objects
from .lib import lib_ast  # noqa: F401
from .lib import load  # noqa: F401
from .lib import load_lib  # noqa: F401
from .registry import NetworkRegistry  # noqa: F401

sys.path.append(str(Path(__file__)))

logger.start()


def module_property(func: Any) -> None:
    """Decorator to turn module functions into properties.
    Function names must be prefixed with an underscore."""
    module = sys.modules[func.__module__]

    def base_getattr(name: str) -> None:
        raise AttributeError(f"module '{module.__name__}' has no attribute '{name}'")

    old_getattr = getattr(module, "__getattr__", base_getattr)

    def new_getattr(name: str) -> Any:
        if f"_{name}" == func.__name__:
            return func()
        else:
            return old_getattr(name)

    module.__getattr__ = new_getattr  # type: ignore
    return func


@module_property
def _networks() -> NetworkRegistry:
    return NetworkRegistry()



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import KeysView
from typing import Optional

# third party
from pydantic import BaseModel


class RouteUpdate(BaseModel):
    source_node_uid: str
    source_node_url: Optional[str]
    private: bool = False
    autodetect: bool = False

    # allows splatting with **payload
    def keys(self) -> KeysView[str]:
        return self.__dict__.keys()

    # allows splatting with **payload
    def __getitem__(self, key: str) -> Any:
        return self.__dict__.__getitem__(key)

    def validate(self) -> bool:
        if self.autodetect:
            return True
        elif self.source_node_url:
            return True
        raise Exception("source_node_url must be supplied if autodetect=False")



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

# third party
from nacl.signing import VerifyKey
from pydantic import BaseSettings
import redis

# syft absolute
import syft as sy

# relative
from .abstract_ledger_store import AbstractDataSubjectLedger
from .abstract_ledger_store import AbstractLedgerStore


class DictLedgerStore(AbstractLedgerStore):
    def __init__(self, *args: Tuple[Any, ...], **kwargs: Any) -> None:
        self.kv_store: Dict[VerifyKey, AbstractDataSubjectLedger] = {}

    def get(self, key: VerifyKey) -> AbstractDataSubjectLedger:
        return self.kv_store[key]

    def set(self, key: VerifyKey, value: AbstractDataSubjectLedger) -> None:
        self.kv_store[key] = value


class RedisLedgerStore(AbstractLedgerStore):
    def __init__(self, settings: Optional[BaseSettings] = None) -> None:

        if settings is None:
            raise Exception("RedisStore requires Settings")
        self.settings = settings
        try:
            self.redis: redis.client.Redis = redis.Redis(
                host=settings.REDIS_HOST,
                port=self.settings.REDIS_PORT,
                db=self.settings.REDIS_LEDGER_DB_ID,
            )
        except Exception as e:
            print("failed to load redis", e)
            raise e

    def get(self, key: VerifyKey) -> AbstractDataSubjectLedger:
        try:
            key_str = bytes(key).hex()
            buf = self.redis.get(key_str)
            if buf is None:
                raise KeyError()
            return sy.deserialize(buf, from_bytes=True)
        except Exception as e:
            print(f"Failed to get ledger from database. {e}")
            raise e

    def set(self, key: VerifyKey, value: AbstractDataSubjectLedger) -> None:
        try:
            key_str = bytes(key).hex()
            buf = sy.serialize(value, to_bytes=True)
            self.redis.set(key_str, buf)
        except Exception as e:
            print(f"Failed to set ledger to database. {e}")
            raise e



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/exception_action.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n4proto/core/node/common/action/exception_action.proto\x12\x1dsyft.core.node.common.service\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xc9\x01\n\x10\x45xceptionMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x37\n\x18msg_id_causing_exception\x18\x03 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x16\n\x0e\x65xception_type\x18\x04 \x01(\t\x12\x15\n\rexception_msg\x18\x05 \x01(\tb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.node.common.action.exception_action_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _EXCEPTIONMESSAGE._serialized_start = 156
    _EXCEPTIONMESSAGE._serialized_end = 357
# @@protoc_insertion_point(module_scope)



# stdlib
from copy import deepcopy
from typing import Any
from typing import Collection
from typing import Iterable
from typing import List
from typing import Optional
from typing import cast

# third party
from pydantic import BaseSettings
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

# syft absolute
import syft as sy

# relative
from ....common.uid import UID
from ....node.common.node_table.bin_obj_dataset import BinObjDataset
from ....store import ObjectStore
from ....store.proxy_dataset import ProxyDataset
from ....store.store_interface import StoreKey
from ....store.storeable_object import StorableObject
from ..node_table.bin_obj_metadata import ObjectMetadata


class DictStore(ObjectStore):
    def __init__(self, db: Session, settings: BaseSettings) -> None:
        self.db = db
        self.settings = settings
        self.kv_store: dict[UID, Any] = {}

    def get_or_none(
        self, key: UID, proxy_only: bool = False
    ) -> Optional[StorableObject]:
        try:
            return self.get(key=key, proxy_only=proxy_only)
        except KeyError as e:  # noqa: F841
            return None

    def get_objects_of_type(self, obj_type: type) -> Iterable[StorableObject]:
        # raise NotImplementedError("get_objects_of_type")
        # return [obj for obj in self.values() if isinstance(obj.data, obj_type)]
        return self.values()

    def __sizeof__(self) -> int:
        return self.values().__sizeof__()

    def __str__(self) -> str:
        return f"{type(self)}"

    def __len__(self) -> int:
        return len(self.kv_store.keys())

    def keys(self) -> Collection[UID]:
        return self.kv_store.keys()

    def values(self) -> List[StorableObject]:
        keys = self.kv_store.keys()
        # this is bad we need to decouple getting the data from the search
        all_values = []
        for key in keys:
            all_values.append(self.get(key=key))

        return all_values

    def __contains__(self, key: StoreKey) -> bool:
        _, key_uid = self.key_to_str_and_uid(key=key)
        return key_uid in self.kv_store

    def resolve_proxy_object(self, obj: Any) -> Any:
        raise Exception(
            f"Proxy Objects {type(obj)} should not be in Nodes with DictStore"
        )

    def __getitem__(self, key: StoreKey) -> StorableObject:
        raise Exception("obj = store[key] not allowed because additional args required")

    def get(self, key: StoreKey, proxy_only: bool = False) -> StorableObject:
        key_str, key_uid = self.key_to_str_and_uid(key=key)
        try:
            local_session = sessionmaker(bind=self.db)()
            store_obj = self.kv_store[key_uid]

            # serialized contents
            if isinstance(store_obj, bytes):
                try:
                    de = sy.deserialize(store_obj, from_bytes=True)
                    obj = de
                except Exception as e:
                    raise Exception(f"Failed to deserialize obj at key {key_str}. {e}")
            else:
                # not serialized
                try:
                    obj = deepcopy(store_obj)
                except Exception as e:
                    raise Exception(
                        f"DictStore should not contain unpickleable objects. {e}"
                    )

            if id(obj) == id(store_obj):
                raise Exception("Objects must use deepcopy or mutation can occur")

            if isinstance(obj, ProxyDataset):
                obj = self.resolve_proxy_object(obj=obj)

            obj_metadata = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).first()
            )

            if obj is None or obj_metadata is None:
                raise KeyError(f"Object not found! for UID: {key_str}")

            obj = StorableObject(
                id=key_uid,
                data=obj.data,
                description=obj_metadata.description,
                tags=obj_metadata.tags,
                read_permissions=dict(
                    sy.deserialize(
                        bytes.fromhex(obj_metadata.read_permissions), from_bytes=True
                    )
                ),
                search_permissions=dict(
                    sy.deserialize(
                        bytes.fromhex(obj_metadata.search_permissions), from_bytes=True
                    )
                ),
                write_permissions=dict(
                    sy.deserialize(
                        bytes.fromhex(obj_metadata.write_permissions), from_bytes=True
                    )
                ),
            )
            local_session.close()
            return obj
        except Exception as e:
            print(f"Cant get object {str(key_str)}", e)
            raise KeyError(f"Object not found! for UID: {str(key_str)}")

    def is_dataset(self, key: StoreKey) -> bool:
        key_str, _ = self.key_to_str_and_uid(key=key)
        local_session = sessionmaker(bind=self.db)()
        is_dataset_obj = (
            local_session.query(BinObjDataset).filter_by(obj=key_str).exists()
        )
        is_dataset_obj = local_session.query(is_dataset_obj).scalar()
        local_session.close()
        return is_dataset_obj

    def __setitem__(self, key: StoreKey, value: StorableObject) -> None:
        self.set(key=key, value=value)

    def set(self, key: StoreKey, value: StorableObject) -> None:
        key_str, key_uid = self.key_to_str_and_uid(key=key)
        try:
            store_obj = deepcopy(value)
            self.kv_store[key_uid] = store_obj
            if id(value) == id(store_obj):
                raise Exception("Objects must use deepcopy or mutation can occur")
        except Exception as e:
            # if we get a pickling error from deepcopy we can juse use sy.serialize
            # in theory not having to do this unless necessary means it should be a
            # little faster and less memory intensive
            if "Pickling" in str(e):
                try:
                    # deepcopy falls back to pickle and some objects can't be pickled
                    self.kv_store[key_uid] = sy.serialize(value, to_bytes=True)
                except Exception as ex:
                    raise Exception(f"Failed to serialize object {type(value)}. {ex}")
            else:
                raise Exception(f"Failed to save object {type(value)}. {e}")

        create_metadata = True
        local_session = sessionmaker(bind=self.db)()
        try:
            # use existing metadata row to prevent more than 1
            metadata_obj = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).all()[-1]
            )
            create_metadata = False
        except Exception:
            pass
            # no metadata row exists lets insert one
            metadata_obj = ObjectMetadata()

        metadata_obj.obj = key_str
        metadata_obj.tags = value.tags
        metadata_obj.description = value.description
        metadata_obj.read_permissions = cast(
            bytes,
            sy.serialize(sy.lib.python.Dict(value.read_permissions), to_bytes=True),
        ).hex()
        metadata_obj.search_permissions = cast(
            bytes,
            sy.serialize(sy.lib.python.Dict(value.search_permissions), to_bytes=True),
        ).hex()
        metadata_obj.write_permissions = cast(
            bytes,
            sy.serialize(sy.lib.python.Dict(value.write_permissions), to_bytes=True),
        ).hex()

        if create_metadata:
            local_session.add(metadata_obj)
        local_session.commit()
        local_session.close()

    def delete(self, key: UID) -> None:
        key_str, key_uid = self.key_to_str_and_uid(key=key)
        try:
            del self.kv_store[key_uid]
            local_session = sessionmaker(bind=self.db)()
            metadata_to_delete = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).first()
            )
            local_session.delete(metadata_to_delete)
            local_session.commit()
            local_session.close()
        except Exception as e:
            print(f"{type(self)} Exception in __delitem__ error {key_str}. {e}")

    def clear(self) -> None:
        self.kv_store = {}
        local_session = sessionmaker(bind=self.db)()
        local_session.query(ObjectMetadata).delete()
        local_session.commit()
        local_session.close()

    def __repr__(self) -> str:
        return f"{type(self)}"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/io/connection.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1eproto/core/io/connection.proto\x12\x0csyft.core.io\x1a\x1bproto/core/io/address.proto">\n\x17VirtualServerConnection\x12#\n\x04node\x18\x01 \x01(\x0b\x32\x15.syft.core.io.Address"P\n\x17VirtualClientConnection\x12\x35\n\x06server\x18\x01 \x01(\x0b\x32%.syft.core.io.VirtualServerConnectionb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.io.connection_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _VIRTUALSERVERCONNECTION._serialized_start = 77
    _VIRTUALSERVERCONNECTION._serialized_end = 139
    _VIRTUALCLIENTCONNECTION._serialized_start = 141
    _VIRTUALCLIENTCONNECTION._serialized_end = 221
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey

# relative
from .....common.message import ImmediateSyftMessageWithReply
from ....domain_interface import DomainInterface
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...exceptions import UserNotFoundError
from ...node_table.utils import model_to_json
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..success_resp_message import SuccessResponseMessage
from .user_messages import CreateUserMessage
from .user_messages import DeleteUserMessage
from .user_messages import GetCandidatesMessage
from .user_messages import GetCandidatesResponse
from .user_messages import GetUserMessage
from .user_messages import GetUserResponse
from .user_messages import GetUsersMessage
from .user_messages import GetUsersResponse
from .user_messages import ProcessUserCandidateMessage
from .user_messages import SearchUsersMessage
from .user_messages import SearchUsersResponse
from .user_messages import UpdateUserMessage


def create_user_msg(
    msg: CreateUserMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check if node requires daa document
    if node.setup.first(domain_name=node.name).daa and not msg.daa_pdf:
        raise AuthorizationError(
            message="You can't apply a new User without a DAA document!"
        )

    # Check if email/password fields are empty
    if not msg.email or not msg.password:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (email/password)!"
        )

    # Check if this email was already registered
    try:
        node.users.first(email=msg.email)
        # If the email has already been registered, raise exception
        raise AuthorizationError(
            message="You can't create a new User using this email!"
        )
    except UserNotFoundError:
        # If email not registered, a new user can be created.
        pass

    app_id = node.users.create_user_application(
        name=msg.name,
        email=msg.email,
        password=msg.password,
        daa_pdf=msg.daa_pdf,
        institution=msg.institution,
        website=msg.website,
        budget=msg.budget,
    )

    user_role_id = -1
    try:
        user_role_id = node.users.role(verify_key=verify_key).id
    except Exception as e:
        print("verify_key not in db", e)

    if node.roles.can_create_users(role_id=user_role_id):
        node.users.process_user_application(
            candidate_id=app_id, status="accepted", verify_key=verify_key
        )

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="User created successfully!",
    )


def accept_or_deny_candidate(
    msg: ProcessUserCandidateMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    if True:  # node.users.can_create_users(verify_key=verify_key):
        node.users.process_user_application(
            candidate_id=msg.candidate_id, status=msg.status, verify_key=verify_key
        )
    else:
        raise AuthorizationError(
            message="You're not allowed to create a new User using this email!"
        )

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="User application processed successfully!",
    )


def update_user_msg(
    msg: UpdateUserMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    _valid_parameters = (
        msg.email
        or (msg.password and msg.new_password)
        or msg.role
        or msg.groups
        or msg.name
        or msg.budget
    )
    _allowed = msg.user_id == 0 or node.users.can_create_users(verify_key=verify_key)
    # Change own information
    if msg.user_id == 0:
        msg.user_id = int(node.users.get_user(verify_key).id)  # type: ignore

    _valid_user = node.users.contain(id=msg.user_id)

    if not _valid_parameters:
        raise MissingRequestKeyError(
            "Missing json fields ( email,password,role,groups, name )"
        )

    if not _allowed:
        raise AuthorizationError("You're not allowed to change other user data!")

    if not _valid_user:
        raise UserNotFoundError

    if msg.institution:
        node.users.set(user_id=str(msg.user_id), institution=msg.institution)

    if msg.website:
        node.users.set(user_id=str(msg.user_id), website=msg.website)

    if msg.budget:
        node.users.set(user_id=str(msg.user_id), budget=msg.budget)

    # Change Email Request
    if msg.email:
        node.users.set(user_id=str(msg.user_id), email=msg.email)

    # Change Password Request
    if msg.password and msg.new_password:
        node.users.change_password(
            user_id=str(msg.user_id),
            current_pwd=msg.password,
            new_pwd=msg.new_password,
        )

    # Change Name Request
    if msg.name:
        node.users.set(user_id=str(msg.user_id), name=msg.name)

    # Change Role Request
    if msg.role:
        target_user = node.users.first(id=msg.user_id)
        _allowed = (
            msg.role != node.roles.owner_role.name  # Target Role != Owner
            and target_user.role
            != node.roles.owner_role.id  # Target User Role != Owner
            and node.users.can_create_users(verify_key=verify_key)  # Key Permissions
        )

        # If all premises were respected
        if _allowed:
            new_role_id = node.roles.first(name=msg.role).id
            node.users.set(user_id=msg.user_id, role=new_role_id)  # type: ignore
        elif target_user.role == node.roles.owner_role.id:
            raise AuthorizationError("You're not allowed to change Owner user roles!")
        else:
            raise AuthorizationError("You're not allowed to change User roles!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="User updated successfully!",
    )


def get_user_msg(
    msg: GetUserMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetUserResponse:
    # Check key permissions
    _allowed = node.users.can_triage_requests(verify_key=verify_key)
    if not _allowed:
        raise AuthorizationError(
            "get_user_msg You're not allowed to get User information!"
        )
    else:
        # Extract User Columns
        user = node.users.first(id=msg.user_id)
        _msg = model_to_json(user)

        # Use role name instead of role ID.
        _msg["role"] = node.roles.first(id=_msg["role"]).name

        # Remove private key
        del _msg["private_key"]

        # Get budget spent
        _msg["budget_spent"] = node.users.get_budget_for_user(
            verify_key=VerifyKey(user.verify_key.encode("utf-8"), encoder=HexEncoder)
        )

    return GetUserResponse(
        address=msg.reply_to,
        content=_msg,
    )


def get_all_users_msg(
    msg: GetUsersMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetUsersResponse:
    # Check key permissions
    _allowed = node.users.can_triage_requests(verify_key=verify_key)
    if not _allowed:
        raise AuthorizationError(
            "get_all_users_msg You're not allowed to get User information!"
        )
    else:
        # Get All Users
        users = node.users.all()
        _msg = []
        for user in users:
            _user_json = model_to_json(user)
            # Use role name instead of role ID.
            _user_json["role"] = node.roles.first(id=_user_json["role"]).name

            # Remove private key
            del _user_json["private_key"]

            # Remaining Budget
            # TODO:
            # Rename it from budget_spent to remaining budget
            _user_json["budget_spent"] = node.users.get_budget_for_user(  # type: ignore
                verify_key=VerifyKey(
                    user.verify_key.encode("utf-8"), encoder=HexEncoder
                ),
            )
            _msg.append(_user_json)

    return GetUsersResponse(
        address=msg.reply_to,
        content=_msg,
    )


def get_applicant_users(
    msg: GetCandidatesMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetCandidatesResponse:
    # Check key permissions
    _allowed = node.users.can_triage_requests(verify_key=verify_key)
    if not _allowed:
        raise AuthorizationError(
            "get_applicant_users You're not allowed to get User information!"
        )
    else:
        # Get All Users
        users = node.users.get_all_applicant()
        _msg = []
        _user_json = {}
        for user in users:
            _user_json = model_to_json(user)
            if user.daa_pdf:
                _user_json["daa_pdf"] = user.daa_pdf
            _msg.append(_user_json)

    return GetCandidatesResponse(
        address=msg.reply_to,
        content=_msg,
    )


def del_user_msg(
    msg: DeleteUserMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:

    _target_user = node.users.first(id=msg.user_id)
    _not_owner = (
        node.roles.first(id=_target_user.role).name != node.roles.owner_role.name
    )

    _allowed = (
        node.users.can_create_users(verify_key=verify_key)  # Key Permission
        and _not_owner  # Target user isn't the node owner
    )
    if _allowed:
        node.users.delete(id=msg.user_id)
    else:
        raise AuthorizationError("You're not allowed to delete this user information!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="User deleted successfully!",
    )


def search_users_msg(
    msg: SearchUsersMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SearchUsersResponse:
    user_parameters = {
        "email": msg.email,
        "role": node.roles.first(name=msg.role).id,
    }

    filtered_parameters = filter(
        lambda key: user_parameters[key], user_parameters.keys()
    )
    user_parameters = {key: user_parameters[key] for key in filtered_parameters}

    _allowed = node.users.can_triage_requests(verify_key=verify_key)

    if _allowed:
        try:
            users = node.users.query(**user_parameters)
            _msg = [model_to_json(user) for user in users]
        except UserNotFoundError:
            _msg = []
    else:
        raise AuthorizationError(
            "search_users_msg You're not allowed to get User information!"
        )

    return SearchUsersResponse(
        address=msg.reply_to,
        content=_msg,
    )


class UserManagerService(ImmediateNodeServiceWithReply):
    INPUT_TYPE = Union[
        Type[CreateUserMessage],
        Type[UpdateUserMessage],
        Type[GetUserMessage],
        Type[GetUsersMessage],
        Type[DeleteUserMessage],
        Type[SearchUsersMessage],
        Type[GetCandidatesMessage],
        Type[ProcessUserCandidateMessage],
    ]

    INPUT_MESSAGES = Union[
        CreateUserMessage,
        UpdateUserMessage,
        GetUserMessage,
        GetUsersMessage,
        DeleteUserMessage,
        SearchUsersMessage,
    ]

    OUTPUT_MESSAGES = Union[
        SuccessResponseMessage,
        GetUserResponse,
        GetUsersResponse,
        GetCandidatesResponse,
        SearchUsersResponse,
    ]

    msg_handler_map: Dict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
        CreateUserMessage: create_user_msg,
        UpdateUserMessage: update_user_msg,
        GetUserMessage: get_user_msg,
        GetUsersMessage: get_all_users_msg,
        DeleteUserMessage: del_user_msg,
        SearchUsersMessage: search_users_msg,
        GetCandidatesMessage: get_applicant_users,
        ProcessUserCandidateMessage: accept_or_deny_candidate,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: INPUT_MESSAGES,
        verify_key: VerifyKey,
    ) -> OUTPUT_MESSAGES:
        reply = UserManagerService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

        return reply

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateUserMessage,
            UpdateUserMessage,
            GetUserMessage,
            GetUsersMessage,
            DeleteUserMessage,
            SearchUsersMessage,
            GetCandidatesMessage,
            ProcessUserCandidateMessage,
        ]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/role_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\'proto/grid/messages/role_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a\x1bproto/lib/python/dict.proto"\xcb\x03\n\x11\x43reateRoleMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1e\n\x16\x63\x61n_make_data_requests\x18\x04 \x01(\x08\x12 \n\x18\x63\x61n_triage_data_requests\x18\x05 \x01(\x08\x12!\n\x19\x63\x61n_manage_privacy_budget\x18\x06 \x01(\x08\x12\x18\n\x10\x63\x61n_create_users\x18\x07 \x01(\x08\x12\x18\n\x10\x63\x61n_manage_users\x18\x08 \x01(\x08\x12\x16\n\x0e\x63\x61n_edit_roles\x18\t \x01(\x08\x12!\n\x19\x63\x61n_manage_infrastructure\x18\n \x01(\x08\x12\x17\n\x0f\x63\x61n_upload_data\x18\x0b \x01(\x08\x12!\n\x19\x63\x61n_upload_legal_document\x18\x0c \x01(\x08\x12 \n\x18\x63\x61n_edit_domain_settings\x18\r \x01(\x08\x12\'\n\x08reply_to\x18\x0e \x01(\x0b\x32\x15.syft.core.io.Address"\x99\x01\n\x0eGetRoleMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07role_id\x18\x03 \x01(\x05\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x88\x01\n\x0fGetRoleResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x02 \x01(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x0fGetRolesMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x10GetRolesResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x02 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\xdc\x03\n\x11UpdateRoleMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1e\n\x16\x63\x61n_make_data_requests\x18\x04 \x01(\x08\x12 \n\x18\x63\x61n_triage_data_requests\x18\x05 \x01(\x08\x12!\n\x19\x63\x61n_manage_privacy_budget\x18\x06 \x01(\x08\x12\x18\n\x10\x63\x61n_create_users\x18\x07 \x01(\x08\x12\x18\n\x10\x63\x61n_manage_users\x18\x08 \x01(\x08\x12\x16\n\x0e\x63\x61n_edit_roles\x18\t \x01(\x08\x12!\n\x19\x63\x61n_manage_infrastructure\x18\n \x01(\x08\x12\x17\n\x0f\x63\x61n_upload_data\x18\x0b \x01(\x08\x12!\n\x19\x63\x61n_upload_legal_document\x18\x0c \x01(\x08\x12 \n\x18\x63\x61n_edit_domain_settings\x18\r \x01(\x08\x12\x0f\n\x07role_id\x18\x0e \x01(\x05\x12\'\n\x08reply_to\x18\x0f \x01(\x0b\x32\x15.syft.core.io.Address"\x9c\x01\n\x11\x44\x65leteRoleMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07role_id\x18\x03 \x01(\x05\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.role_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEROLEMESSAGE._serialized_start = 161
    _CREATEROLEMESSAGE._serialized_end = 620
    _GETROLEMESSAGE._serialized_start = 623
    _GETROLEMESSAGE._serialized_end = 776
    _GETROLERESPONSE._serialized_start = 779
    _GETROLERESPONSE._serialized_end = 915
    _GETROLESMESSAGE._serialized_start = 918
    _GETROLESMESSAGE._serialized_end = 1055
    _GETROLESRESPONSE._serialized_start = 1058
    _GETROLESRESPONSE._serialized_end = 1195
    _UPDATEROLEMESSAGE._serialized_start = 1198
    _UPDATEROLEMESSAGE._serialized_end = 1674
    _DELETEROLEMESSAGE._serialized_start = 1677
    _DELETEROLEMESSAGE._serialized_end = 1833
# @@protoc_insertion_point(module_scope)



"""fucntion used to calculate log of given tensor."""
# future
from __future__ import annotations

# stdlib
from typing import TYPE_CHECKING
from typing import Union

# relative
from .exp import exp

if TYPE_CHECKING:
    # relative
    from ...tensor.smpc.mpc_tensor import MPCTensor


def log(
    data: MPCTensor, iterations: int = 2, exp_iterations: int = 8
) -> Union[float, MPCTensor]:
    """Approximates the natural logarithm using 8th order modified Householder iterations.
        Recall that Householder method is an algorithm to solve a non linear equation f(x) = 0.
        Here  f: x -> 1 - C * exp(-x)  with C = data
        Iterations are computed by:
            y_0 = some constant
            h = 1 - data * exp(-y_n)
            y_{n+1} = y_n - h * (1 + h / 2 + h^2 / 3 + h^3 / 6 + h^4 / 5 + h^5 / 7)

    Args:
        data: tensor whose log has to be calculated
        iterations (int): number of iterations for 6th order modified
            Householder approximation.
        exp_iterations (int): number of iterations for limit approximation of exp

    Returns:
        MPCTensor: Calculated log value of given tensor
    Ref: https://github.com/LaRiffle/approximate-models
    """
    y = data / 31 + 1.59 - 20 * exp((-2 * data - 1.4), iterations=exp_iterations)

    # 6th order Householder iterations
    for i in range(iterations):
        h = [1 - data * exp((-1 * y), iterations=exp_iterations)]
        for j in range(1, 5):
            h.append(h[-1] * h[0])

        y -= h[0] * (1 + h[0] / 2 + h[1] / 3 + h[2] / 4 + h[3] / 5 + h[4] / 6)

    return y



# stdlib
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

# relative
from ...adp.data_subject_ledger import DataSubjectLedger
from ...common.serde.serializable import serializable
from ..autodp.gamma_tensor import GammaTensor
from ..autodp.phi_tensor import PhiTensor
from ..tensor import Tensor
from .layers.base import Layer
from .loss import BinaryCrossEntropy
from .loss import Loss
from .optimizers import Adamax
from .optimizers import Optimizer


@serializable(recursive_serde=True)
class Model:
    __attr_allowlist__ = [
        "layers",
        "loss",
        "optimizer",
        "aggregated_loss",
    ]

    def __init__(self, layers: Optional[List[Type[Layer]]] = None):
        self.layers: List = [] if layers is None else layers
        self.aggregated_loss: float = 0.0

    def publish(
        self,
        deduct_epsilon_for_user: Callable,
        get_budget_for_user: Callable,
        ledger: DataSubjectLedger,
        sigma: float,
    ) -> Dict:
        print("Publish Model Weights")

        parameters = {}
        for i, layer in enumerate(self.layers):
            print("Layer", str(layer))

            print("Before  Publish")
            for param in layer.params:  # type: ignore
                print(param.shape, end=" ")
            print()
            if hasattr(layer, "params"):
                parameters[str(layer) + str(i)] = [
                    param.publish(
                        deduct_epsilon_for_user=deduct_epsilon_for_user,
                        get_budget_for_user=get_budget_for_user,
                        ledger=ledger,
                        sigma=sigma,
                    )
                    if isinstance(param, (GammaTensor))
                    else param
                    for param in layer.params  # type: ignore
                ]
                print("After  Publish")
                for param in parameters[str(layer) + str(i)]:
                    print(param.shape, end=" ")
                print()

        parameters["loss"] = self.aggregated_loss  # type: ignore

        return parameters

    def replace_weights(self, published_weights: dict) -> None:
        """
        For when you want to use published weights downloaded from a domain
        """
        for i, layer in enumerate(self.layers):
            params = published_weights[str(layer) + str(i)]
            if len(params) > 0:
                layer.params = params  # type: ignore

    def add(self, layer: Layer) -> None:
        if isinstance(layer, Layer):
            self.layers.append(layer)
        else:
            raise TypeError("PySyft doesn't recognize this kind of layer.")

    def initialize_weights(
        self,
        loss: Loss = BinaryCrossEntropy(),
        optimizer: Optimizer = Adamax(),
    ) -> None:
        self.layers[0].first_layer = True

        prev_layer = None
        for _, layer in enumerate(self.layers):
            layer.connect_to(prev_layer=prev_layer)
            prev_layer = layer

        self.loss = loss
        self.optimizer = optimizer

    def fit(
        self,
        X: Union[GammaTensor, PhiTensor, Tensor],
        Y: Union[GammaTensor, PhiTensor, Tensor],
        max_iter: int = 100,
        batch_size: int = 4,
        shuffle: bool = True,
        validation_split: float = 0.0,
        validation_data: Optional[Tuple] = None,
    ) -> None:
        print("Started Training")

        if isinstance(X, Tensor):
            X = X.child
        if isinstance(Y, Tensor):
            Y = Y.child

        # prepare data
        train_X = X  # .astype(get_dtype()) if np.issubdtype(np.float64, X.dtype) else X
        train_Y = Y  # .astype(get_dtype()) if np.issubdtype(np.float64, Y.dtype) else Y

        if 1.0 > validation_split > 0.0:
            split = int(train_Y.shape[0] * validation_split)
            valid_X, valid_Y = train_X[-split:], train_Y[-split:]
            train_X, train_Y = train_X[:-split], train_Y[:-split]
        elif validation_data is not None:
            valid_X, valid_Y = validation_data
        else:
            valid_X, valid_Y = None, None

        iter_idx = 0
        while iter_idx < max_iter:
            iter_idx += 1

            # train
            train_losses = []
            # train_predicts, train_targets =  [], []
            for b in range(train_Y.shape[0] // batch_size):
                batch_begin = b * batch_size
                batch_end = batch_begin + batch_size
                x_batch = train_X[batch_begin:batch_end]
                y_batch = train_Y[batch_begin:batch_end]

                # forward propagation
                y_pred = self.predict(x_batch)

                # backward propagation
                next_grad = self.loss.backward(y_pred, y_batch)
                for layer in self.layers[::-1]:
                    print("Backward layer", str(layer))
                    next_grad = layer.backward(next_grad)

                # update parameters
                self.optimizer.update(self.layers)

                # got loss and predict
                train_losses.append(self.loss.forward(y_pred, y_batch))

            if valid_X is not None and valid_Y is not None:
                valid_losses, valid_predicts, valid_targets = [], [], []  # type: ignore

                for b in range(valid_X.shape[0] // batch_size):
                    batch_begin = b * batch_size
                    batch_end = batch_begin + batch_size
                    x_val_batch = valid_X[batch_begin:batch_end]
                    y_val_batch = valid_Y[batch_begin:batch_end]

                    # forward propagation
                    y_val_pred = self.predict(x_val_batch)

                    # got loss and predict
                    valid_losses.append(self.loss.forward(y_val_pred, y_val_batch))
                    valid_predicts.extend(y_val_pred.child)
                    valid_targets.extend(y_val_batch.child)

    def step(
        self,
        x_batch: Union[GammaTensor, PhiTensor, Tensor],
        y_batch: Union[GammaTensor, PhiTensor, Tensor],
    ) -> None:

        x_batch = x_batch.child if isinstance(x_batch, Tensor) else x_batch
        y_batch = y_batch.child if isinstance(y_batch, Tensor) else y_batch

        print(type(x_batch), type(y_batch))
        print(x_batch.shape, y_batch.shape)

        # forward propagation
        y_pred = self.predict(x_batch)

        print("Predictions:", y_pred.child.argmax())
        print("Actual Value:", y_batch.child.shape)

        # backward propagation
        next_grad = self.loss.backward(y_pred, y_batch)
        for layer in self.layers[::-1]:
            print("Backward layer", str(layer))
            next_grad = layer.backward(next_grad)

        # update parameters
        print("Updating optimizer")
        self.optimizer.update(self.layers)

        # got loss and predict
        print("Predicting loss")
        loss = self.loss.forward(y_pred, y_batch)

        curr_loss = float(loss.child)
        self.aggregated_loss += curr_loss

        print("Loss:", self.aggregated_loss)

    def predict(
        self, X: Union[GammaTensor, PhiTensor, Tensor]
    ) -> Union[GammaTensor, PhiTensor, Tensor]:
        """Calculate an output Y for the given input X."""
        x_next = X
        for layer in self.layers:
            print("Forward layer", str(layer))
            x_next = layer.forward(x_next)
        y_pred = x_next
        return y_pred



# relative
from .group.group import LocationGroup
from .group.registry import RegistryBackedLocationGroup
from .group.subscription import SubscriptionBackedLocationGroup
from .location import Location
from .specific import SpecificLocation

__all__ = [
    "Location",
    "SpecificLocation",
    "LocationGroup",
    "RegistryBackedLocationGroup",
    "SubscriptionBackedLocationGroup",
]



"""Utility functions for approximation functions."""
# future
from __future__ import annotations

# stdlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # relative
    from ...tensor.smpc.mpc_tensor import MPCTensor


def sign(data: MPCTensor) -> MPCTensor:
    """Calculate sign of given tensor.

    Args:
        data: tensor whose sign has to be determined

    Returns:
        MPCTensor: tensor with the determined sign
    """
    return (data > 0) + (data < 0) * (-1)


def modulus(data: MPCTensor) -> MPCTensor:
    """Calculation of modulus for a given tensor.

    Args:
        data(MPCTensor): tensor whose modulus has to be calculated

    Returns:
        MPCTensor: the required modulus
    """
    return sign(data) * data



# stdlib
from typing import Any
from typing import Callable as CallableT
import warnings

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft

# relative
from ....util import aggressive_set_attr
from .capnp import CAPNP_REGISTRY

module_type = type(syft)


# this will overwrite the ._sy_serializable_wrapper_type with an auto generated
# wrapper which will basically just hold the object being wrapped.
def GenerateWrapper(
    wrapped_type: type,
    import_path: str,
    protobuf_scheme: GeneratedProtocolMessageType,
    type_object2proto: CallableT,
    type_proto2object: CallableT,
) -> None:
    @serializable()
    class Wrapper:
        def __init__(self, value: object):
            self.obj = value

        def _object2proto(self) -> Any:
            return type_object2proto(self.obj)

        @staticmethod
        def _proto2object(proto: Any) -> Any:
            return type_proto2object(proto)

        @staticmethod
        def get_protobuf_schema() -> GeneratedProtocolMessageType:
            return protobuf_scheme

        def upcast(self) -> Any:
            return self.obj

        @staticmethod
        def wrapped_type() -> type:
            return wrapped_type

    # TODO: refactor like proxy class to get correct name
    # WARNING: Changing this can break the Wrapper lookup during deserialize
    module_parts = import_path.split(".")
    klass = module_parts.pop()
    Wrapper.__name__ = f"{klass}Wrapper"
    Wrapper.__module__ = f"syft.wrappers.{'.'.join(module_parts)}"
    # create a fake module `wrappers` under `syft`
    if "wrappers" not in syft.__dict__:
        syft.__dict__["wrappers"] = module_type(name="wrappers")
    # for each part of the path, create a fake module and add it to it's parent
    parent = syft.__dict__["wrappers"]
    for n in module_parts:
        if n not in parent.__dict__:
            parent.__dict__[n] = module_type(name=n)
        parent = parent.__dict__[n]
    # finally add our wrapper class to the end of the path
    parent.__dict__[Wrapper.__name__] = Wrapper

    aggressive_set_attr(
        obj=wrapped_type, name="_sy_serializable_wrapper_type", attr=Wrapper
    )


def GenerateProtobufWrapper(
    cls_pb: GeneratedProtocolMessageType, import_path: str
) -> None:
    def object2proto(obj: Any) -> Any:
        return obj

    def proto2object(proto: Any) -> Any:
        return proto

    GenerateWrapper(
        wrapped_type=cls_pb,
        import_path=import_path,
        protobuf_scheme=cls_pb,
        type_object2proto=object2proto,
        type_proto2object=proto2object,
    )


def serializable(
    generate_wrapper: bool = False,
    protobuf_object: bool = False,
    recursive_serde: bool = False,
    capnp_bytes: bool = False,
) -> Any:
    def rs_decorator(cls: Any) -> Any:
        # relative
        from .recursive import rs_get_protobuf_schema
        from .recursive import rs_object2proto
        from .recursive import rs_proto2object

        if not hasattr(cls, "__attr_allowlist__"):
            warnings.warn(
                f"__attr_allowlist__ not defined for type {cls.__name__},"
                " even if it uses recursive serde, defaulting on the empty list."
            )
            cls.__attr_allowlist__ = []

        if not hasattr(cls, "__serde_overrides__"):
            cls.__serde_overrides__ = {}

        cls._object2proto = rs_object2proto
        cls._proto2object = staticmethod(rs_proto2object)
        cls.get_protobuf_schema = staticmethod(rs_get_protobuf_schema)
        return cls

    def serializable_decorator(cls: Any) -> Any:
        protobuf_schema = cls.get_protobuf_schema()
        # overloading a protobuf by adding multiple classes and we will check the
        # obj_type string later to dispatch to the correct one
        if hasattr(protobuf_schema, "schema2type"):
            if isinstance(protobuf_schema.schema2type, list):
                protobuf_schema.schema2type.append(cls)
            else:
                protobuf_schema.schema2type = [protobuf_schema.schema2type, cls]
        else:
            protobuf_schema.schema2type = cls
        return cls

    def capnp_decorator(cls: Any) -> Any:
        # register deserialize with the capnp registry
        CAPNP_REGISTRY[cls.__name__] = cls._bytes2object
        return cls

    if capnp_bytes:
        return capnp_decorator

    if generate_wrapper:
        return GenerateWrapper

    if recursive_serde:
        return rs_decorator

    return serializable_decorator
